# distribution_auto_fitting
# Author: Yan Fu
# -*- coding: utf-8 -*-
# Date: 07/03/2019

import os
import warnings
from tkinter import filedialog, Tk, StringVar

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.stats as st
from typing import Union
from scipy import interpolate


# Create models from data
def fit(
    data,
    bins: int = 200,
    ax=None,
    distribution_list: Union[list, int] = None,
    suppress_print: bool = False,
    fmt_str: str = "{:20.20}{:60.60}",
):
    """
    Model data by finding best fit distribution to data
    :param data:
    :param bins:
    :param ax:
    :param distribution_list:
    :param suppress_print:
    :param fmt_str:
    :return list_fitted_distribution:
    :return list_fitted_distribution_params:
    :return list_fitted_distribution_sse:
    """

    # Get histogram of original data
    y, x = np.histogram(data, bins=bins, density=True)
    x = (x + np.roll(x, -1))[:-1] / 2.0
    # ax.plot(x, y)
    # ax.fill_between(x, y, color="black")

    def cprint(*args):
        if suppress_print:
            return 0
        else:
            print(*args)

    if ax:
        ax.hist(data, bins="auto", density=True, color="black")
        ax.set_xlim(
            min(*x) - (max(*x) - min(*x)) * 0.01,
            max(*x) + (max(*x) - min(*x)) * 0.01,
            auto=False,
        )
        ax.set_ylim(
            min(*y) - (max(*y) - min(*y)) * 0.01,
            max(*y) + (max(*y) - min(*y)) * 0.01,
            auto=False,
        )

    if distribution_list == 0 or distribution_list == None:
        # full list of available distribution functions
        distribution_list = [
            st.alpha,
            st.anglit,
            st.arcsine,
            st.beta,
            st.betaprime,
            st.bradford,
            st.burr,
            st.cauchy,
            st.chi,
            st.chi2,
            st.cosine,
            st.dgamma,
            st.dweibull,
            st.erlang,
            st.expon,
            st.exponnorm,
            st.exponweib,
            st.exponpow,
            st.f,
            st.fatiguelife,
            st.fisk,
            st.foldcauchy,
            st.foldnorm,
            st.frechet_r,
            st.frechet_l,
            st.genlogistic,
            st.genpareto,
            st.gennorm,
            st.genexpon,
            st.genextreme,
            st.gausshyper,
            st.gamma,
            st.gengamma,
            st.genhalflogistic,
            st.gilbrat,
            st.gompertz,
            st.gumbel_r,
            st.gumbel_l,
            st.halfcauchy,
            st.halflogistic,
            st.halfnorm,
            st.halfgennorm,
            st.hypsecant,
            st.invgamma,
            st.invgauss,
            st.invweibull,
            st.johnsonsb,
            st.johnsonsu,
            st.ksone,
            st.kstwobign,
            st.laplace,
            st.levy,
            st.levy_l,
            st.levy_stable,
            st.logistic,
            st.loggamma,
            st.loglaplace,
            st.lognorm,
            st.lomax,
            st.maxwell,
            st.mielke,
            st.nakagami,
            st.ncx2,
            st.ncf,
            st.nct,
            st.norm,
            st.pareto,
            st.pearson3,
            st.powerlaw,
            st.powerlognorm,
            st.powernorm,
            st.rdist,
            st.reciprocal,
            st.rayleigh,
            st.rice,
            st.recipinvgauss,
            st.semicircular,
            st.t,
            st.triang,
            st.truncexpon,
            st.truncnorm,
            st.tukeylambda,
            st.uniform,
            st.vonmises,
            st.vonmises_line,
            st.wald,
            st.weibull_min,
            st.weibull_max,
            st.wrapcauchy,
        ]
    elif distribution_list == 1:
        # https://blog.cloudera.com/blog/2015/12/common-probability-distributions-the-data-scientists-crib-sheet/
        distribution_list = [
            st.uniform,
            st.bernoulli,
            st.hypergeom,
            st.binom,
            st.geom,
            st.gumbel_r,
            st.gumbel_l,
            st.nbinom,
            st.poisson,
            st.expon,
            st.lognorm,
            st.weibull_max,
            st.weibull_min,
            st.dweibull,
            st.norm,
            st.chi,
            st.gamma,
            st.beta,
        ]

    # Best holders
    best_distribution = st.norm
    best_params = (0.0, 1.0)
    best_sse = np.inf

    # Estimate distribution parameters from data

    cprint(
        fmt_str.format(
            "Distribution", "Loss (Residual sum of squares) and distribution parameters"
        )
    )
    list_fitted_distribution = []
    list_fitted_distribution_sse = []
    list_fitted_distribution_params = []

    for distribution in distribution_list:

        # Try to fit the distribution
        try:
            # Ignore warnings from data that can't be fit
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore")

                # fit dist to data
                params = distribution.fit(data)

                # Separate parts of parameters
                arg = params[:-2]
                loc = params[-2]
                scale = params[-1]

                # Calculate fitted PDF and error with fit in distribution
                pdf = distribution.pdf(x, loc=loc, scale=scale, *arg)
                sse = np.sum(np.power(y - pdf, 2.0))

                # if axis pass in add to plot
                try:
                    if ax:
                        ax.plot(x, pdf, label=distribution.name)
                except Exception:
                    pass

                # identify if this distribution is better
                if best_sse > sse > 0:
                    best_distribution = distribution
                    best_params = params
                    best_sse = sse

                params_ = ["{:10.2f}".format(i) for i in params]

                cprint(
                    fmt_str.format(
                        distribution.name,
                        "{:10.5E} - [{}]".format(sse, ",".join(params_)),
                    )
                )

                list_fitted_distribution.append(distribution)
                list_fitted_distribution_sse.append(sse)
                list_fitted_distribution_params.append(params)

        except Exception:
            pass

    return (
        list_fitted_distribution,
        list_fitted_distribution_params,
        list_fitted_distribution_sse,
    )


def cdf_to_samples(
    cdf_x: Union[list, np.ndarray],
    cdf_y: Union[list, np.ndarray],
    n_samples: int = 10000,
):
    """Produce a number of samples for a given cumulative density function (CDF) in x and y. The samples will be
    randomly selected.

    :param cdf_x: the x value of the CDF
    :param cdf_y: the y value of the CDF
    :param n_samples: number of samples desired
    :return samples: produced samples
    """

    cdf_func = interpolate.interp1d(cdf_y, cdf_x, kind="linear")
    cdf_pts = np.random.uniform(high=np.max(cdf_y), low=np.min(cdf_y), size=n_samples)

    samples = cdf_func(cdf_pts).flatten()

    return samples


def pdf_to_samples(
    pdf_x: Union[list, np.ndarray],
    pdf_y: Union[list, np.ndarray],
    n_samples: int = 10000,
):
    """Produce a number of samples for a given probability density function (PDF) in x and y. The samples will be
    randomly selected.

    :param pdf_x: the x value of the PDF
    :param pdf_y: the y value of the PDF
    :param n_samples: number of samples desired
    :return samples: produced samples
    """

    cdf_x = pdf_x
    cdf_y = np.cumsum(pdf_y)

    samples = cdf_to_samples(cdf_x, cdf_y, n_samples)

    return samples


def auto_fit():
    # INPUTS

    # Note 1: *.csv must contain no headers and numbers only.

    # Define data type
    # 0 - single column containing sampled values
    # 1 - two columns with 1st column containing values and second column probability
    # 2 - two columns with 1st column containing values and second column cumulative probability

    print("This function will be depreciated, use `sfeprapy distfit`.")

    data_type = input(
        "Define Data Type\n0 - single column containing sampled values\n1 - two columns with 1st column containing values and second column probability\n2 - two columns with 1st column containing values and second column cumulative probability\nEnter your value and press Enter: "
    )
    data_type = int(data_type)

    # Distribution type or list
    # list - will only fit to the provided list of distributions (scipy.stats.*)
    # 0 - fit to all available distributions
    # 1 - fit to common distribution types
    distribution_list = input(
        "Distribution Type\n0 - fit to all available distributions\n1 - fit to common distribution types\nEnter your value and press Enter: "
    )
    distribution_list = int(distribution_list)

    # Select data csv file
    root = Tk()
    root.withdraw()
    folder_path = StringVar()
    path_input_file_csv = filedialog.askopenfile(
        title="Select data csv file", filetypes=[("csv", [".csv"])]
    )
    folder_path.set(path_input_file_csv)
    root.update()

    dir_work = os.path.dirname(path_input_file_csv.name)
    print(dir_work)

    # Load data
    if data_type == 0:
        data_csv = pd.read_csv(path_input_file_csv, header=None, dtype=float)
        samples = data_csv.values.flatten()
    elif data_type == 1:
        data_csv = pd.read_csv(path_input_file_csv, header=None, dtype=float)
        samples = pdf_to_samples(
            pdf_x=data_csv[0].values.flatten(),
            pdf_y=data_csv[1].values.flatten(),
            n_samples=10000,
        )
    elif data_type == 2:
        data_csv = pd.read_csv(path_input_file_csv, header=None, dtype=float)
        #     print(data_csv[1])
        samples = cdf_to_samples(
            cdf_x=data_csv[0].values.flatten(),
            cdf_y=data_csv[1].values.flatten(),
            n_samples=10000,
        )
    else:
        samples = np.nan

    # FITTING DISTRIBUTIONS TO DATA

    fig_fitting, ax_fitting = plt.subplots(figsize=(3.94 * 2.5, 2.76 * 2.5))
    list_dist, list_params, list_sse = fit(
        samples, ax=ax_fitting, distribution_list=distribution_list
    )

    # FINDING THE BEST FIT

    list_dist = np.asarray(list_dist)[np.argsort(list_sse)]
    list_params = np.asarray(list_params)[np.argsort(list_sse)]
    list_sse = np.asarray(list_sse)[np.argsort(list_sse)]
    print(
        "{:30.30}{:60.60}".format(
            "Distribution (sorted)",
            "Loss (Residual sum of squares) and distribution parameters",
        )
    )
    for i, v in enumerate(list_dist):
        print(
            "{:30.30}{:10.5E} - [{}]".format(
                v.name,
                list_sse[i],
                ", ".join(["{:10.2f}".format(i) for j in list_params[i]]),
            )
        )

    dist_best = list_dist[0]
    params_best = list_params[0]

    # PRODUCE FIGURES

    print("Figures saved at", dir_work)

    ax_fitting.set_xlabel("Sample value")
    ax_fitting.set_ylabel("PDF")
    ax_fitting.legend().set_visible(True)
    fig_fitting.savefig(os.path.join(dir_work, "fitting.png"))

    cdf_x_sampled = np.sort(samples)
    cdf_y_sampled = np.linspace(0, 1, len(cdf_x_sampled))
    cdf_y_fitted = np.linspace(0, 1, 1000)
    cdf_x_fitted = dist_best.ppf(cdf_y_fitted, *params_best)

    fig_results, ax_results = plt.subplots(figsize=(3.94 * 2.5, 2.76 * 2.5))
    ax_results.hist(samples, bins="auto", density=True, color="black")
    ax_results.plot(cdf_x_sampled, cdf_y_sampled, label="Sampled", color="black")
    ax_results.plot(cdf_x_fitted, cdf_y_fitted, label="Fitted", color="red")
    ax_results.set_xlabel("Sample values")
    ax_results.set_ylabel("CDF")
    ax_results.legend().set_visible(True)
    fig_results.savefig(os.path.join(dir_work, "results.png"))

    input("Press any key to finish.")


def auto_fit_2(
    data_type: int, distribution_list: Union[int, list], data: Union[str, pd.DataFrame]
):
    """
    :param data_type:
    :param distribution_list:   a list of distribution function names as defined in scipy.stats; or integers:
                                0   single column containing sampled values.
                                1   two columns with 1st column containing values and second column probability.
                                2   two columns with 1st column containing values and second column cumulative
                                    probability.
    :param data:    a csv file path string or a pandas.DataFrame object; or
                    0   fit to all available distributions
                    1   fit to common distribution types
    :return:
    """

    if isinstance(data, str):
        data = os.path.realpath(data)
        dir_work = os.path.dirname(data)
        data_csv = pd.read_csv(data, header=None, dtype=float)
    elif isinstance(data, pd.DataFrame):
        data_csv = data
        dir_work = os.getcwd()
    else:
        raise ValueError("Unknown data type.")

    # Load data
    if data_type == 0:
        samples = data_csv.values.flatten()
    elif data_type == 1:
        data_csv = pd.read_csv(data, header=None, dtype=float)
        samples = pdf_to_samples(
            pdf_x=data_csv[0].values.flatten(),
            pdf_y=data_csv[1].values.flatten(),
            n_samples=10000,
        )
    elif data_type == 2:
        data_csv = pd.read_csv(data, header=None, dtype=float)
        #     print(data_csv[1])
        samples = cdf_to_samples(
            cdf_x=data_csv[0].values.flatten(),
            cdf_y=data_csv[1].values.flatten(),
            n_samples=10000,
        )
    else:
        samples = np.nan

    # FITTING DISTRIBUTIONS TO DATA

    fig_fitting, ax_fitting = plt.subplots(figsize=(3.94 * 2.5, 2.76 * 2.5))
    list_dist, list_params, list_sse = fit(
        samples,
        ax=ax_fitting,
        distribution_list=distribution_list,
        fmt_str="{:30.30}{}",
    )

    # FINDING THE BEST FIT

    list_dist = np.asarray(list_dist)[np.argsort(list_sse)]
    list_params = np.asarray(list_params)[np.argsort(list_sse)]
    list_sse = np.asarray(list_sse)[np.argsort(list_sse)]

    print(
        "\n{:30.30}{}".format(
            "Distribution (sorted)",
            "Loss (Residual sum of squares) and distribution parameters",
        )
    )
    for i, v in enumerate(list_dist):
        dist_name = v.name
        sse = list_sse[i]
        dist_params = ", ".join(["{:10.2f}".format(j) for j in list_params[i]])
        print(f"{dist_name:30.30}{sse:10.5E} - [{dist_params}]")

    dist_best = list_dist[0]
    params_best = list_params[0]

    # PRODUCE FIGURES

    ax_fitting.set_xlabel("Sample value")
    ax_fitting.set_ylabel("PDF")
    ax_fitting.legend().set_visible(True)
    fig_fitting.savefig(os.path.join(dir_work, "distfit_fitting.png"))

    cdf_x_sampled = np.sort(samples)
    cdf_y_sampled = np.linspace(0, 1, len(cdf_x_sampled))
    cdf_y_fitted = np.linspace(0, 1, 1000)
    cdf_x_fitted = dist_best.ppf(cdf_y_fitted, *params_best)

    fig_results, ax_results = plt.subplots(figsize=(3.94 * 2.5, 2.76 * 2.5))

    ax_results.hist(samples, bins="auto", density=True, color="black")
    ax_results.plot(cdf_x_sampled, cdf_y_sampled, label="Sampled", color="black")
    ax_results.plot(cdf_x_fitted, cdf_y_fitted, label="Fitted", color="red")
    ax_results.set_xlabel("Sample values")
    ax_results.set_ylabel("CDF")
    ax_results.legend().set_visible(True)
    fig_results.savefig(os.path.join(dir_work, "distfit_results.png"))



"""
VERSION IDENTIFICATION RULES DOCUMENTED IN PEP 440 ARE FOLLOWED.

Version scheme
==============

Distributions are identified by a public version identifier which supports all defined version comparison operations

The version scheme is used both to describe the distribution version provided by a particular distribution archive, as
well as to place constraints on the version of dependencies needed in order to build or run the software.

Public version identifiers
--------------------------

The canonical public version identifiers MUST comply with the following scheme:

`[N!]N(.N)*[{a|b|rc}N][.postN][.devN]`

Public version identifiers MUST NOT include leading or trailing whitespace.

Public version identifiers MUST be unique within a given distribution.

See also Appendix B : Parsing version strings with regular expressions which provides a regular expression to check
strict conformance with the canonical format, as well as a more permissive regular expression accepting inputs that may
require subsequent normalization.

Public version identifiers are separated into up to five segments:

    - Epoch segment: N!
    - Release segment: N(.N)*
    - Pre-release segment: {a|b|rc}N
    - Post-release segment: .postN
    - Development release segment: .devN

"""

__version__ = "0.6.9"


def check_pip_upgrade():

    # Parse the latest version string
    import subprocess
    from subprocess import STDOUT, check_output

    try:
        output = check_output("pip search --version sfeprapy", stderr=STDOUT, timeout=5)
    except (subprocess.TimeoutExpired, subprocess.CalledProcessError) as e:
        return

    # extract the version number string
    import re

    v = re.findall(r"sfeprapy[\s]*\([\d.]+\)", str(output))[0]
    v = re.findall(r"[\d.]+", str(v))[0]

    # check if upgrade required
    from packaging import version

    is_new_version_available = version.parse(v) > version.parse(__version__)

    # raise message if upgrade is needed
    if is_new_version_available:
        print(
            "New SfePrapy version is available, use `pip install sfeprapy --upgrade` to install the latest version."
        )
        print(f"Current: {__version__}\nLatest: {v}\n\n")


if __name__ == "__main__":
    import re

    def is_canonical(version):
        return (
            re.match(
                r"^([1-9][0-9]*!)?(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))*((a|b|rc)(0|[1-9][0-9]*))?(\.post(0|[1-9][0-9]*))?(\.dev(0|[1-9][0-9]*))?$",
                version,
            )
            is not None
        )

    assert is_canonical(__version__)

    check_pip_upgrade()



def fire(time, growth_factor, cap_hrr=0, cap_hrr_to_time=0):
    if isinstance(growth_factor, str):
        growth_factor = str.lower(growth_factor)
        growth_dict = {
            "slow": 0.0029,
            "medium": 0.0117,
            "fast": 0.0469,
            "ultra-fast": 0.1876,
        }
        try:
            growth_factor = growth_dict[growth_factor]
        except KeyError:
            err_msg = "{} should be one of the following if not a number: {}"
            err_msg = err_msg.format(
                "growth_factor", ", ".join(list(growth_dict.keys()))
            )
            raise ValueError(err_msg)

    heat_release_rate = growth_factor * time ** 2 * 1000

    # cap hrr
    if cap_hrr:
        heat_release_rate[heat_release_rate > cap_hrr] = cap_hrr

    if cap_hrr_to_time:
        heat_release_rate[time > cap_hrr_to_time] = -1
        heat_release_rate[heat_release_rate == -1] = max(heat_release_rate)

    return heat_release_rate



# -*- coding: utf-8 -*-
import matplotlib

from sfeprapy.mcs1.mcs1_obj import MonteCarlo

matplotlib.use("agg")


def main(path_master_csv: str = None):
    MC = MonteCarlo()
    MC.select_input_file(path_master_csv)
    MC.make_input_param()
    MC.make_mc_params()
    MC.run_mc()
    MC.out_combined_ky()
    MC.out_combined_T()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


def save_figure(fp_mcs0_out: str):
    import os
    import numpy as np
    import pandas as pd
    import plotly.io
    import plotly.graph_objects as go

    mcs_out = pd.read_csv(fp_mcs0_out)

    # A helper function to produce (x, y) line plot based upon samples
    def cdf_xy(v, xlim, bin_width=0.1, weights=None):
        edges = np.arange(*xlim, bin_width)
        x = (edges[1:] + edges[:-1]) / 2
        y_pdf = np.histogram(v, bins=edges, weights=weights)[0] / len(v)
        y_cdf = np.cumsum(y_pdf)
        return x, y_cdf

    # Settings
    # fig_size = np.array([3.5, 3.5]) * 1  # in inch
    fig_x_limit = (0, 180)
    fig_y_limit = (0, 1)
    bin_width = 0.5

    # Process time equivalence value, obtain `probability_weight` and `n_simulations`
    list_case_name = sorted(list(set(mcs_out["case_name"].values)))
    list_t_eq = list()
    list_weight = list()
    list_n_simulation = list()
    for case_name in list_case_name:
        teq = np.asarray(
            mcs_out[mcs_out["case_name"] == case_name][
                "solver_time_equivalence_solved"
            ].values,
            float,
        )
        teq[teq == np.inf] = np.max(teq[teq != np.inf])
        teq[teq == -np.inf] = np.min(teq[teq != -np.inf])
        teq = teq[~np.isnan(teq)]
        teq[
            teq > 18000.0
        ] = 18000.0  # limit maximum time equivalence plot value to 5 hours
        list_t_eq.append(teq / 60.0)
        list_weight.append(
            np.average(
                mcs_out[mcs_out["case_name"] == case_name]["probability_weight"].values
            )
        )
        list_n_simulation.append(
            np.average(
                mcs_out[mcs_out["case_name"] == case_name]["n_simulations"].values
            )
        )

    # Time equivalence samples -> x, y of cumulative density function
    xlim = (0, np.max([np.max(v) for v in list_t_eq]) + bin_width)
    x = np.arange(*xlim, bin_width)
    list_cdf_t_eq_y = [
        cdf_xy(t_eq, xlim, bin_width, weights=None)[1] for t_eq in list_t_eq
    ]

    # Combined time equivalence cdf
    cdf_t_eq_y_combined = np.array(
        [list_weight[i] * v for i, v in enumerate(list_cdf_t_eq_y)]
    )
    cdf_t_eq_y_combined = np.sum(cdf_t_eq_y_combined, axis=0)

    # Plot figure
    fig = go.Figure()

    # add combined time equivalence to the plot
    if len(list_case_name) > 1 and np.max(cdf_t_eq_y_combined) > 0.9:
        fig.add_trace(
            go.Scatter(x=x, y=cdf_t_eq_y_combined, mode="lines", name="Combined")
        )

    # add individual time equivalence to the plot
    for i, t_eq in enumerate(list_cdf_t_eq_y):
        fig.add_trace(go.Scatter(x=x, y=t_eq, mode="lines", name=list_case_name[i]))

    fig.update_layout(
        autosize=True,
        paper_bgcolor="White",
        plot_bgcolor="White",
        xaxis=dict(
            title="Equivalent of time exposure [minute]",
            dtick=30,
            range=fig_x_limit,
            showline=True,
            linewidth=1,
            linecolor="black",
            mirror=True,
            visible=True,
            showgrid=True,
            gridcolor="Black",
            gridwidth=1,
            ticks="outside",
            zeroline=False,
        ),
        yaxis=dict(
            title="CDF",
            dtick=0.2,
            range=fig_y_limit,
            showline=True,
            linewidth=1,
            linecolor="black",
            mirror=True,
            visible=True,
            showgrid=True,
            gridcolor="Black",
            gridwidth=1,
            ticks="outside",
            zeroline=False,
        ),
        legend=dict(
            x=0.98,
            xanchor="right",
            y=0.02,
            yanchor="bottom",
            traceorder="normal",
            font=dict(family="sans-serif", color="black"),
            bgcolor="White",
            bordercolor="Black",
            borderwidth=1,
        ),
    )

    config = {
        "scrollZoom": False,
        "displayModeBar": True,
        "editable": True,
        "showLink": False,
        "displaylogo": False,
    }

    plotly.io.write_html(
        fig,
        file=os.path.join(os.path.dirname(fp_mcs0_out), "mcs.out.html"),
        auto_open=True,
        config=config,
    )


def main(fp_mcs_in: str, n_threads: int = None):

    import os
    import warnings
    from sfeprapy.func.mcs_obj import MCS
    from sfeprapy.mcs0.mcs0_calc import teq_main, teq_main_wrapper, mcs_out_post
    from sfeprapy.func.mcs_gen import main as gen

    warnings.filterwarnings("ignore")

    fp_mcs_in = os.path.realpath(fp_mcs_in)

    mcs = MCS()
    mcs.define_problem(fp_mcs_in)
    mcs.define_stochastic_parameter_generator(gen)
    mcs.define_calculation_routine(teq_main, teq_main_wrapper, mcs_out_post)

    try:
        if n_threads:
            mcs.config = (
                dict(n_threads=n_threads)
                if mcs.config
                else mcs.config.update(dict(n_threads=n_threads))
            )
    except KeyError:
        pass

    mcs.run_mcs()


if __name__ == "__main__":
    print("Use `sfeprapy` CLI, `sfeprapy.mcs0:__main__` is depreciated on 22 Oct 2019.")



if __name__ == "__main__":
    import sys
    import warnings
    import sfeprapy.mcs1.mcs1 as mc1

    warnings.filterwarnings("ignore")

    if len(sys.argv) > 1:
        mc1.main(sys.argv[1])
    else:
        mc1.main()



# -*- coding: utf-8 -*-
import numpy as np


def protected_steel_eurocode(
    time,
    temperature_ambient,
    rho_steel,
    area_steel_section,
    k_protection,
    rho_protection,
    c_protection,
    thickness_protection,
    perimeter_protected,
    terminate_when_cooling=False,
    terminate_max_temperature=np.inf,
):
    """
    SI UNITS!
    This function calculate the temperature curve of protected steel section based on BS EN 1993-1-2:2005, Section 4
    . Ambient (fire) time-temperature data must be given, as well as the parameters specified below.
    :param time: ndarray, [s], time evolution.
    :param temperature_ambient: ndarray, [K], imposed temperature evolution.
    :param rho_steel: float, [kg/m3], steel density.
    :param area_steel_section: float, [m2], steel cross section area.
    :param k_protection: float, [K/kg/m], protection thermal conductivity.
    :param rho_protection: float, [kg/m3], protection density.
    :param c_protection: float, [J/K/kg], protection thermal capacity.
    :param thickness_protection: float, [m], protection thickness.
    :param perimeter_protected: float, [m], protected perimeter.
    :param terminate_when_cooling: bool, [-], if True then terminate and return values when first peak steel
    temperature is observed.
    :return temperature_steel: ndarray, [K], is calculated steel temperature.
    """
    # todo: 4.2.5.2 (2) - thermal properties for the insulation material
    # todo: revise BS EN 1993-1-2:2005, Clauses 4.2.5.2

    # BS EN 1993-1-2:2005, 3.4.1.2
    # return self.__make_property("c")

    def c_steel_T(temperature):

        temperature -= 273.15
        if temperature < 20:
            # warnings.warn('Temperature ({:.1f} °C) is below 20 °C'.format(temperature))
            return (
                425 + 0.773 * 20 - 1.69e-3 * np.power(20, 2) + 2.22e-6 * np.power(20, 3)
            )
        if 20 <= temperature < 600:
            return (
                425
                + 0.773 * temperature
                - 1.69e-3 * np.power(temperature, 2)
                + 2.22e-6 * np.power(temperature, 3)
            )
        elif 600 <= temperature < 735:
            return 666 + 13002 / (738 - temperature)
        elif 735 <= temperature < 900:
            return 545 + 17820 / (temperature - 731)
        elif 900 <= temperature <= 1200:
            return 650
        elif temperature > 1200:
            # warnings.warn('Temperature ({:.1f} °C) is greater than 1200 °C'.format(temperature))
            return 650
        else:
            # warnings.warn('Temperature ({:.1f} °C) is outside bound.'.format(temperature))
            return 0

    V = area_steel_section
    rho_a = rho_steel
    lambda_p = k_protection
    rho_p = rho_protection
    d_p = thickness_protection
    A_p = perimeter_protected
    c_p = c_protection

    temperature_steel = time * 0.0
    temperature_rate_steel = time * 0.0
    specific_heat_steel = time * 0.0

    # Check time step <= 30 seconds. [BS EN 1993-1-2:2005, Clauses 4.2.5.2 (3)]
    # time_change = gradient(time)
    # if np.max(time_change) > 30.:
    # raise ValueError("Time step needs to be less than 30s: {0}".format(np.max(time)))

    flag_heating_started = False

    temperature_steel[0] = temperature_ambient[
        0
    ]  # initially, steel temperature is equal to ambient
    temperature_ambient_ = iter(temperature_ambient)  # skip the first item
    next(temperature_ambient_)  # skip the first item
    for i, T_g in enumerate(temperature_ambient_):
        i += 1  # actual index since the first item had been skipped.
        try:
            specific_heat_steel[i] = c_steel_T(temperature_steel[i - 1])
        except ValueError:
            specific_heat_steel[i] = specific_heat_steel[i - 1]
            # print(temperature_steel[i-1])

        # Steel temperature equations are from [BS EN 1993-1-2:2005, Clauses 4.2.5.2, Eq. 4.27]
        phi = (c_p * rho_p / specific_heat_steel[i] / rho_a) * d_p * A_p / V

        a = (lambda_p * A_p / V) / (d_p * specific_heat_steel[i] * rho_a)
        b = (T_g - temperature_steel[i - 1]) / (1.0 + phi / 3.0)
        c = (np.exp(phi / 10.0) - 1.0) * (T_g - temperature_ambient[i - 1])
        d = time[i] - time[i - 1]

        temperature_rate_steel[i] = (
            a * b * d - c
        ) / d  # deviated from e4.27, converted to rate [s-1]

        temperature_steel[i] = temperature_steel[i - 1] + temperature_rate_steel[i] * d

        if (temperature_rate_steel[i] > 0 and flag_heating_started is False) and time[
            i
        ] > 1800:
            flag_heating_started = True

        # Terminate steel temperature calculation if necessary
        if (
            terminate_when_cooling
            and flag_heating_started
            and temperature_rate_steel[i] < 0
        ):
            break
        elif flag_heating_started and terminate_max_temperature < temperature_steel[i]:
            break

        # NOTE: Steel temperature can be in cooling phase at the beginning of calculation, even the ambient temperature
        #       (fire) is hot. This is
        #       due to the factor 'phi' which intends to address the energy locked within the protection layer.
        #       The steel temperature is forced to be increased or remain as previous when ambient temperature and
        #       its previous temperature are all higher than the current calculated temperature.
        #       A better implementation is perhaps to use a 1-D heat transfer model.

        # DEPRECIATED 26 MAR 2019
        # if temperature_steel[i] < temperature_steel[i-1] or temperature_steel[i] < temperature_ambient[i]:
        #     temperature_rate_steel[i] = 0
        #     temperature_steel[i] = temperature_steel[i-1]

    return temperature_steel


def protected_steel_eurocode_max_temperature(
    time,
    temperature_ambient,
    rho_steel,
    area_steel_section,
    k_protection,
    rho_protection,
    c_protection,
    thickness_protection,
    perimeter_protected,
    terminate_check_wait_time=3600,
    terminate_max_temperature=np.inf,
):
    """
    LIMITATIONS:
    Constant time interval throughout
    Only one maxima

    SI UNITS!
    This function calculate the temperature curve of protected steel section based on BS EN 1993-1-2:2005, Section 4
    . Ambient (fire) time-temperature data must be given, as well as the parameters specified below.
    :param time:                    {ndarray} [s]
    :param temperature_ambient:     {ndarray} [K]
    :param rho_steel:               {float} [kg/m3]
    :param area_steel_section:      {float} [m2]
    :param k_protection:            {float} [K/kg/m]
    :param rho_protection:          {float} [kg/m3]
    :param c_protection:            {float} [J/K/kg]
    :param thickness_protection:    {float} [m]
    :param perimeter_protected:     {float} [m]
                                                            temperature is observed.
    :return time:                   {ndarray, float} [s]
    :return temperature_steel:      {ndarray, float} [K]
    :return data_all:               {Dict} [-]
    """
    # todo: 4.2.5.2 (2) - thermal properties for the insulation material
    # todo: revise BS EN 1993-1-2:2005, Clauses 4.2.5.2

    def c_steel_T(temperature):

        temperature -= 273.15
        if temperature < 20:
            # warnings.warn('Temperature ({:.1f} °C) is below 20 °C'.format(temperature))
            return (
                425 + 0.773 * 20 - 1.69e-3 * np.power(20, 2) + 2.22e-6 * np.power(20, 3)
            )
        if 20 <= temperature < 600:
            return (
                425
                + 0.773 * temperature
                - 1.69e-3 * np.power(temperature, 2)
                + 2.22e-6 * np.power(temperature, 3)
            )
        elif 600 <= temperature < 735:
            return 666 + 13002 / (738 - temperature)
        elif 735 <= temperature < 900:
            return 545 + 17820 / (temperature - 731)
        elif 900 <= temperature <= 1200:
            return 650
        elif temperature > 1200:
            # warnings.warn('Temperature ({:.1f} °C) is greater than 1200 °C'.format(temperature))
            return 650
        else:
            # warnings.warn('Temperature ({:.1f} °C) is outside bound.'.format(temperature))
            return 0

    V = area_steel_section
    rho_a = rho_steel
    lambda_p = k_protection
    rho_p = rho_protection
    d_p = thickness_protection
    A_p = perimeter_protected
    c_p = c_protection

    # temperature_steel = time * 0.
    # temperature_rate_steel = time * 0.
    # specific_heat_steel = time * 0.

    # Check time step <= 30 seconds. [BS EN 1993-1-2:2005, Clauses 4.2.5.2 (3)]
    # time_change = gradient(time)
    # if np.max(time_change) > 30.:
    # raise ValueError("Time step needs to be less than 30s: {0}".format(np.max(time)))

    flag_heating_started = False

    # T_ini = temperature_ambient[0]  # temperature at beginning
    T = temperature_ambient[0]  # current steel temperature
    # T_max = -1  # maximum steel temperature
    # c = c_steel_T(293.15)
    d = time[1] - time[0]

    for i in range(1, len(temperature_ambient)):

        T_g = temperature_ambient[i]

        c_s = c_steel_T(T)

        # Steel temperature equations are from [BS EN 1993-1-2:2005, Clauses 4.2.5.2, Eq. 4.27]
        phi = (c_p * rho_p / c_s / rho_a) * d_p * A_p / V

        a = (lambda_p * A_p / V) / (d_p * c_s * rho_a)
        b = (T_g - T) / (1.0 + phi / 3.0)
        c = (np.exp(phi / 10.0) - 1.0) * (T_g - temperature_ambient[i - 1])

        dT = (a * b * d - c) / d  # deviated from e4.27, converted to rate [s-1]

        T += dT * d

        if not (T >= 0 or T <= 0):
            print("d")

        if not flag_heating_started:
            if time[i] >= terminate_check_wait_time:
                if dT > 0:
                    flag_heating_started = True

        # Terminate early if maximum temperature is reached
        if flag_heating_started and T > terminate_max_temperature:
            break
        elif flag_heating_started and dT < 0:
            T -= dT * d
            break

    return T


if __name__ == "__main__":

    from sfeprapy.func.fire_iso834 import fire
    from sfeprapy.dat.steel_carbon import Thermal
    import numpy as np

    import matplotlib.pyplot as plt

    plt.style.use("seaborn-paper")
    fig, ax = plt.subplots(figsize=(3.94, 2.76))
    ax.set_xlabel("Time [minute]")
    ax.set_ylabel("Temperature [$^{℃}C$]")
    ax.legend().set_visible(True)
    ax.grid(color="k", linestyle="--")

    steel_prop = Thermal()
    c = steel_prop.c()
    rho = 7850
    t = np.arange(0, 181 * 60, 1)
    T = fire(t, 20 + 273.15)
    ax.plot(t / 60, T - 273.15, "k", label="ISO 834 fire")

    list_dp = np.arange(0.002, 0.01 + 0.002, 0.002)

    for d_p in list_dp:
        T_s = protected_steel_eurocode(
            time=t,
            temperature_ambient=T,
            rho_steel=rho,
            area_steel_section=0.017,
            k_protection=0.2,
            rho_protection=800,
            c_protection=1700,
            thickness_protection=d_p,
            perimeter_protected=2.14,
        )
        plt.plot(
            t / 60,
            T_s - 273.15,
            label="Protection thickness {:2.0f} mm".format(d_p * 1000),
        )

    plt.legend(loc=4)
    plt.tight_layout()
    plt.show()



# -*- coding: utf-8 -*-


def __example_config_dict():
    y = dict(n_threads=2)
    return y


def __example_input_dict():
    y = {
        "Standard Case 1": dict(
            case_name="Standard Case 1",
            n_simulations=500,
            probability_weight=1 / 3,
            fire_time_step=30,
            fire_time_duration=18000,
            fire_hrr_density=dict(
                dist="uniform_", lbound=0.25 - 0.001, ubound=0.25 + 0.001
            ),
            fire_load_density=dict(
                dist="gumbel_r_", lbound=10, ubound=1500, mean=420, sd=126
            ),
            fire_spread_speed=dict(dist="uniform_", lbound=0.0035, ubound=0.0190),
            fire_nft_limit=dict(
                dist="norm_", lbound=623.15, ubound=2023.15, mean=1323.15, sd=93
            ),
            fire_combustion_efficiency=dict(dist="uniform_", lbound=0.8, ubound=1.0),
            window_open_fraction=dict(
                dist="lognorm_mod_", ubound=0.9999, lbound=0.0001, mean=0.2, sd=0.2
            ),
            phi_teq=dict(dist="constant_", ubound=1, lbound=1, mean=0, sd=0),
            beam_cross_section_area=0.017,
            beam_position_horizontal=-1,
            beam_position_vertical=3.2,
            beam_rho=7850,
            fire_mode=3,
            fire_gamma_fi_q=1,
            fire_t_alpha=300,
            fire_tlim=0.333,
            protection_c=1700,
            protection_k=0.2,
            protection_protected_perimeter=2.14,
            protection_rho=800,
            room_breadth=16,
            room_depth=31.25,
            room_height=3.3,
            room_wall_thermal_inertia=720,
            solver_temperature_goal=893.15,
            solver_max_iter=20,
            solver_thickness_lbound=0.0001,
            solver_thickness_ubound=0.0500,
            solver_tol=1.0,
            window_height=2.8,
            window_width=72,
            window_open_fraction_permanent=0,
            timber_exposed_area=0,
            timber_charring_rate=0.7,  # mm/min
            timber_hc=13.2,  # MJ/kg
            timber_density=400,  # [kg/m3]
            timber_solver_ilim=20,
            timber_solver_tol=1,
        ),
        "Standard Case 2 (with teq_phi)": dict(
            case_name="Standard Case 2 (with teq_phi)",
            n_simulations=500,
            probability_weight=1 / 3,
            fire_time_step=30,
            fire_time_duration=18000,
            fire_hrr_density=dict(
                dist="uniform_", lbound=0.25 - 0.001, ubound=0.25 + 0.001
            ),
            fire_load_density=dict(
                dist="gumbel_r_", lbound=10, ubound=1500, mean=420, sd=126
            ),
            fire_spread_speed=dict(dist="uniform_", lbound=0.0035, ubound=0.0190),
            fire_nft_limit=dict(
                dist="norm_", lbound=623.15, ubound=2023.15, mean=1323.15, sd=93
            ),
            fire_combustion_efficiency=dict(dist="uniform_", lbound=0.8, ubound=1.0),
            window_open_fraction=dict(
                dist="lognorm_mod_", ubound=0.9999, lbound=0.0001, mean=0.2, sd=0.2
            ),
            phi_teq=dict(dist="lognorm_", ubound=3, lbound=0.00001, mean=1, sd=0.25),
            beam_cross_section_area=0.017,
            beam_position_horizontal=-1,
            beam_position_vertical=3.2,
            beam_rho=7850,
            fire_mode=3,
            fire_gamma_fi_q=1,
            fire_t_alpha=300,
            fire_tlim=0.333,
            protection_c=1700,
            protection_k=0.2,
            protection_protected_perimeter=2.14,
            protection_rho=800,
            room_breadth=16,
            room_depth=31.25,
            room_height=3.3,
            room_wall_thermal_inertia=720,
            solver_temperature_goal=893.15,
            solver_max_iter=20,
            solver_thickness_lbound=0.0001,
            solver_thickness_ubound=0.0500,
            solver_tol=1.0,
            window_height=2.8,
            window_width=72,
            window_open_fraction_permanent=0,
            timber_exposed_area=0,
            timber_charring_rate=0.7,  # [mm/min]
            timber_hc=13.2,  # [MJ/kg]
            timber_density=400,  # [kg/m3]
            timber_solver_ilim=20,
            timber_solver_tol=1,
        ),
        "Standard Case 3 (with timber)": dict(
            case_name="Standard Case 3 (with timber)",
            n_simulations=500,
            probability_weight=1 / 3,
            fire_time_step=30,
            fire_time_duration=18000,
            fire_hrr_density=dict(
                dist="uniform_", lbound=0.25 - 0.001, ubound=0.25 + 0.001
            ),
            fire_load_density=dict(
                dist="gumbel_r_", lbound=10, ubound=1500, mean=420, sd=126
            ),
            fire_spread_speed=dict(dist="uniform_", lbound=0.0035, ubound=0.0190),
            fire_nft_limit=dict(
                dist="norm_", lbound=623.15, ubound=2023.15, mean=1323.15, sd=93
            ),
            fire_combustion_efficiency=dict(dist="uniform_", lbound=0.8, ubound=1.0),
            window_open_fraction=dict(
                dist="lognorm_mod_", ubound=0.9999, lbound=0.0001, mean=0.2, sd=0.2
            ),
            phi_teq=dict(dist="constant_", ubound=1, lbound=1, mean=0, sd=0),
            beam_cross_section_area=0.017,
            beam_position_horizontal=-1,
            beam_position_vertical=3.2,
            beam_rho=7850,
            fire_mode=3,
            fire_gamma_fi_q=1,
            fire_t_alpha=300,
            fire_tlim=0.333,
            protection_c=1700,
            protection_k=0.2,
            protection_protected_perimeter=2.14,
            protection_rho=800,
            room_breadth=16,
            room_depth=31.25,
            room_height=3.3,
            room_wall_thermal_inertia=720,
            solver_temperature_goal=893.15,
            solver_max_iter=20,
            solver_thickness_lbound=0.0001,
            solver_thickness_ubound=0.0500,
            solver_tol=1.0,
            window_height=2.8,
            window_width=72,
            window_open_fraction_permanent=0,
            timber_exposed_area=500,
            timber_charring_rate=0.7,  # mm/min
            timber_hc=13.2,  # MJ/kg
            timber_density=400,  # [kg/m3]
            timber_solver_ilim=20,
            timber_solver_tol=1,
        ),
    }
    return y


def __example_input_csv(x: dict):
    from sfeprapy.func.mcs_gen import dict_flatten
    import pandas as pd

    y = {k: dict_flatten(v) for k, v in x.items()}
    y = pd.DataFrame.from_dict(y, orient="columns")
    y.index.name = "PARAMETERS"
    y = y.to_csv(index=True)
    return y


EXAMPLE_CONFIG_DICT = __example_config_dict()
EXAMPLE_INPUT_DICT = __example_input_dict()
EXAMPLE_INPUT_CSV = __example_input_csv(__example_input_dict())

if __name__ == "__main__":
    print(EXAMPLE_CONFIG_DICT, "\n")
    print(EXAMPLE_INPUT_DICT, "\n")
    print(EXAMPLE_INPUT_CSV, "\n")



# -*- coding: utf-8 -*-
import copy


# [Eq. 11.1 & 11.2]
def _theta_c(Q, r, h):
    """
    :param Q: [W] Rate of heat release from the fire.
    :param r: [m] Radial distance from the centre of the fire plume impingement.
    :param h: [m]
    :return theta_c: [K]
    """
    Q = copy.copy(Q) / 1000.0  # Unit: [W] to [kW]

    # Condition for r/h
    if r / h > 0.18:
        theta_c = 5.38 * (Q / r) ** (2 / 3) / h
    else:
        theta_c = 16.9 * Q ** (2 / 3) / h ** (5 / 3)

    theta_c += 273.15  # Unit: [C] to [K]
    return theta_c


def _U(Q, r, h):
    """[Eq. 11.3 & 11.4]
    :param Q: [W] Rate of heat release from the fire.
    :param r: [m] Radial distance from the centre of the fire plume impingement.
    :param h: [m]
    :return U: [m/s]
    """
    Q = copy.copy(Q) / 1000.0  # Unit: [W] to [kW]

    if r / h > 0.15:
        U = 0.195 * Q ** (1 / 3) * h ** (1 / 2) / r ** (5 / 6)
    else:
        U = 0.96 * Q / h ** (1 / 3)

    return U


def _dT_d_dt(U, T_g, T_d, RTI):
    """[Eq. 11.5]
    :param U: [m/s]
    :param T_g: [K]
    :param T_d: [K]
    :param RTI: [-]
    :return dT_d_dt: [K/s]
    """

    dT_d_dt = U ** (1 / 2) * (T_g - T_d) / RTI

    return dT_d_dt


def _Q(alpha, t):
    """[Eq. 6.1]
    :param alpha:
    :param t:
    :return:
    """
    Q = alpha * t ** 2
    return Q


if __name__ == "__main__":
    import numpy as np

    # INPUTS
    time_start = 0
    time_step = 0.5
    time_end = 5 * 60

    alpha = 0.0117e3  # [W/s2]
    r = 2.75  # Estimation
    h = 2.6
    RTI = 115
    T_d_activation = 273.15 + 68  # [K]

    # CONTAINERS
    time = np.arange(time_start, time_end, time_step)
    Q = time * 0
    T_g = time * 0
    T_d = time * 0

    # INITIAL CONDITIONS
    T_g[0] = 273.15
    T_d[0] = 273.15

    # CALCULATION
    # calculate heat release rate
    Q = _Q(alpha, time)
    # calculate jet speed near sprinkler
    U = _U(Q, r, h)

    # calculate temperature near sprinkler
    T_g = _theta_c(Q, r, h)

    # calculate sprinkler temperature
    iter_time = enumerate(time)
    next(iter_time)
    for i, t in iter_time:
        T_d[i] = T_d[i - 1] + _dT_d_dt(U[i], T_g[i], T_d[i - 1], RTI)

    # RE-EVALUATE FIRE FOR SUPPRESSION ACTIVATION
    Q[T_d >= T_d_activation] = -1
    T_g[T_d >= T_d_activation] = -1
    T_d[T_d >= T_d_activation] = -1
    Q[Q == -1] = np.max(Q)
    T_d[T_d == -1] = np.max(T_d)
    T_g[T_g == -1] = np.max(T_g)

    print(
        "{:25}: {:7.2f} [min]".format(
            "Sprinkler activated at", np.min(time[T_d == np.max(T_d)]) / 60.0
        )
    )
    print("{:25}: {:7.2f} [kW]".format("The maximum HRR is", np.max(Q) / 1.0e3))

    import matplotlib.pyplot as plt

    plt.figure(num=1)
    plt.subplot(211)
    plt.plot(time / 60, Q / 1e6, label="HRR")
    plt.ylabel("Heat Release Rate [MW]")
    plt.subplot(212)
    plt.plot(time / 60, T_g - 273.15, label="Temperature (near field)")
    plt.plot(time / 60, T_d - 273.15, label="Temperature (sprinkler)")
    plt.xlabel("Time [min]")
    plt.ylabel("Temperature [$\degree C$]")
    plt.legend()
    plt.show()



import numpy as np


def fire(time, temperature_ambient):
    time /= 1200.0  # convert from seconds to hours
    temperature_ambient -= 273.15  # convert temperature from kelvin to celcius
    temperature = (
        750 * (1 - np.exp(-3.79553 * np.sqrt(time)))
        + 170.41 * np.sqrt(time)
        + temperature_ambient
    )
    return temperature + 273.15  # convert from celsius to kelvin (SI unit)



# -*- coding: utf-8 -*-
import copy
from typing import Union

import numpy as np


def fire(
    t: np.array,
    fire_load_density_MJm2: float,
    fire_hrr_density_MWm2: float,
    room_length_m: float,
    room_width_m: float,
    fire_spread_rate_ms: float,
    beam_location_height_m: float,
    beam_location_length_m: Union[float, list, np.ndarray],
    fire_nft_limit_c: float,
    opening_fraction: float = 0,
    opening_width_m: float = 0,
    opening_height_m: float = 0,
):
    """
    This function calculates and returns a temperature array representing travelling fire. This function is NOT in SI.
    :param t: in s, is the time array
    :param fire_load_density_MJm2: in MJ/m2, is the fuel density on the floor
    :param fire_hrr_density_MWm2: in MW/m2, is the heat release rate density
    :param room_length_m: in m, is the room length
    :param room_width_m: in m, is the room width
    :param fire_spread_rate_ms: in m/s, is fire spread speed
    :param beam_location_height_m: in m, is the beam lateral distance to fire origin
    :param beam_location_length_m: in m, is the beam height above the floor
    :param fire_nft_limit_c: in deg.C, is the maximum near field temperature
    :param opening_fraction: in -, is the ventilation opening proportion between 0 to 1
    :param opening_width_m: in m, is ventilation opening width
    :param opening_height_m: in m, is ventilation opening height
    :return T_g: in deg.C, is calculated gas temperature
    """

    # re-assign variable names for equation readability
    q_fd = fire_load_density_MJm2
    HRRPUA = fire_hrr_density_MWm2
    s = fire_spread_rate_ms
    h_s = beam_location_height_m
    l_s = beam_location_length_m
    l = room_length_m
    w = room_width_m
    if l < w:
        l += w
        w = l - w
        l -= w

    # work out ventilation conditions

    # a_v = opening_height_m * opening_width_m * opening_fraction
    # Qv = 1.75 * a_v * np.sqrt(opening_height_m)

    # workout burning time etc.
    t_burn = max([q_fd / HRRPUA, 900.0])
    t_decay = max([t_burn, l / s])
    t_lim = min([t_burn, l / s])

    # reduce resolution to fit time step for t_burn, t_decay, t_lim
    time_interval_s = t[1] - t[0]
    t_decay_ = round(t_decay / time_interval_s, 0) * time_interval_s
    t_lim_ = round(t_lim / time_interval_s, 0) * time_interval_s
    if t_decay_ == t_lim_:
        t_lim_ -= time_interval_s

    # workout the heat release rate ARRAY (corrected with time)
    Q_growth = (HRRPUA * w * s * t) * (t < t_lim_)
    Q_peak = (
        min([HRRPUA * w * s * t_burn, HRRPUA * w * l]) * (t >= t_lim_) * (t <= t_decay_)
    )
    Q_decay = (max(Q_peak) - (t - t_decay_) * w * s * HRRPUA) * (t > t_decay_)
    Q_decay[Q_decay < 0] = 0
    Q = (Q_growth + Q_peak + Q_decay) * 1000.0

    # workout the distance between fire median to the structural element r
    l_fire_front = s * t
    l_fire_front[l_fire_front < 0] = 0
    l_fire_front[l_fire_front > l] = l
    l_fire_end = s * (t - t_lim)
    l_fire_end[l_fire_end < 0] = 0.0
    l_fire_end[l_fire_end > l] = l
    l_fire_median = (l_fire_front + l_fire_end) / 2.0

    # workout the far field temperature of gas T_g
    if isinstance(l_s, float) or isinstance(l_s, int):
        r = np.absolute(l_s - l_fire_median)
        T_g = np.where((r / h_s) > 0.8, (5.38 * np.power(Q / r, 2 / 3) / h_s) + 20.0, 0)
        T_g = np.where(
            (r / h_s) <= 0.8,
            (16.9 * np.power(Q, 2 / 3) / np.power(h_s, 5 / 3)) + 20.0,
            T_g,
        )
        T_g[T_g >= fire_nft_limit_c] = fire_nft_limit_c
        return T_g
    elif isinstance(l_s, np.ndarray) or isinstance(l_s, list):
        l_s_list = copy.copy(l_s)
        T_g_list = list()
        for l_s in l_s_list:
            r = np.absolute(l_s - l_fire_median)
            T_g = np.where(
                (r / h_s) > 0.8, (5.38 * np.power(Q / r, 2 / 3) / h_s) + 20.0, 0
            )
            T_g = np.where(
                (r / h_s) <= 0.8,
                (16.9 * np.power(Q, 2 / 3) / np.power(h_s, 5 / 3)) + 20.0,
                T_g,
            )
            T_g[T_g >= fire_nft_limit_c] = fire_nft_limit_c
            T_g_list.append(T_g)
        return T_g_list
    else:
        raise TypeError('Unknown type of parameter "l_s": {}'.format(type(l_s)))


def fire_backup(
    t: np.ndarray,
    T_0: float,
    q_fd: float,
    hrrpua: float,
    l: float,
    w: float,
    s: float,
    e_h: float,
    e_l: float,
    T_max: float = 1323.15,
):
    """
    :param t: ndarray, [s] An array representing time incorporating 'temperature'.
    :param T_0: float, [K] ,Initial temperature.
    :param q_fd: float, [J/m2], Fire load density.
    :param hrrpua: float, [W/m2], Heat release rate density.
    :param l: float, [m], Compartment length.
    :param w: float, [m], Compartment width.
    :param s: float, [m/s], Fire spread speed.
    :param e_h: float, [m], Vertical distance between element to fuel bed.
    :param e_l: float, [m], Horizontal distance between element to fire front.
    :return temperature: [K] An array representing temperature incorporating 'time'.
    """

    # UNIT CONVERSION TO FIT EQUATIONS
    T_0 -= 273.15
    q_fd /= 1e6
    hrrpua /= 1e6
    T_max -= 273.15

    # workout time step
    time_step = t[1] - t[0]

    # workout burning time etc.
    t_burn = max([q_fd / hrrpua, 900.0])
    t_decay = max([t_burn, l / s])
    t_lim = min([t_burn, l / s])

    # reduce resolution to fit time step for t_burn, t_decay, t_lim
    t_decay_ = round(t_decay / time_step, 0) * time_step
    t_lim_ = round(t_lim / time_step, 0) * time_step
    if t_decay_ == t_lim_:
        t_lim_ -= time_step

    # workout the heat release rate ARRAY (corrected with time)
    Q_growth = (hrrpua * w * s * t) * (t < t_lim_)
    Q_peak = (
        min([hrrpua * w * s * t_burn, hrrpua * w * l]) * (t >= t_lim_) * (t <= t_decay_)
    )
    Q_decay = (max(Q_peak) - (t - t_decay_) * w * s * hrrpua) * (t > t_decay_)
    Q_decay[Q_decay < 0] = 0
    Q = (Q_growth + Q_peak + Q_decay) * 1000.0

    # workout the distance between fire_curve median to the structural element r
    l_fire_front = s * t
    l_fire_front[l_fire_front < 0] = 0.0
    l_fire_front[l_fire_front > l] = l
    l_fire_end = s * (t - t_lim)
    l_fire_end[l_fire_end < 0] = 0.0
    l_fire_end[l_fire_end > l] = l
    l_fire_median = (l_fire_front + l_fire_end) / 2.0
    r = np.absolute(e_l - l_fire_median)
    r[r == 0] = 0.001  # will cause crash if r = 0

    # workout the far field temperature of gas T_g
    T_g1 = (5.38 * np.power(Q / r, 2 / 3) / e_h) * ((r / e_h) > 0.18)
    T_g2 = (16.9 * np.power(Q, 2 / 3) / np.power(e_h, 5 / 3)) * ((r / e_h) <= 0.18)
    T_g = T_g1 + T_g2 + T_0

    T_g[T_g >= T_max] = T_max

    # UNIT CONVERSION TO FIT OUTPUT (SI)
    T_g = T_g + 273.15  # C -> K
    Q *= 10e6  # MJ -> J

    return T_g


def example_plot_interflam():
    time = np.arange(0, 210 * 60, 30)
    list_l = [25, 50, 100, 150]

    import matplotlib.pyplot as plt

    plt.style.use("seaborn-paper")
    fig, ax = plt.subplots(figsize=(3.94, 2.76))
    ax.set_xlabel("Time [minute]")
    ax.set_ylabel("Temperature [$℃$]")

    for length in list_l:
        temperature = fire(
            t=time,
            fire_load_density_MJm2=600,
            fire_hrr_density_MWm2=0.25,
            room_length_m=length,
            room_width_m=16,
            fire_spread_rate_ms=0.012,
            beam_location_height_m=3,
            beam_location_length_m=length / 2,
            fire_nft_limit_c=1050,
        )

        ax.plot(time / 60, temperature, label="Room length {:4.0f} m".format(length))

    ax.legend(loc=4).set_visible(True)
    ax.set_xlim((0, 180))
    ax.set_ylim((0, 1400))
    ax.grid(color="k", linestyle="--")
    plt.tight_layout()
    plt.savefig(fname="fire-travelling.png", dpi=300)


def _test_fire_backup():
    import numpy as np

    time = np.arange(0, 22080, 30)
    list_l = [50, 100, 150]

    import matplotlib.pyplot as plt

    plt.style.use("seaborn-paper")
    fig, ax = plt.subplots(figsize=(3.94, 2.76))
    ax.set_xlabel("Time [minute]")
    ax.set_ylabel("Temperature [$℃$]")

    for l in list_l:
        temperature = fire_backup(
            t=time,
            T_0=293.15,
            q_fd=900e6,
            hrrpua=0.15e6,
            l=l,
            w=17.4,
            s=0.012,
            e_h=3.5,
            e_l=l / 2,
        )
        ax.plot(time / 60, temperature - 273.15)

    ax.legend().set_visible(True)
    ax.set_xlim((0, 120))
    ax.grid(color="k", linestyle="--")
    plt.tight_layout()
    # plt.show()


def _test_fire():
    time = np.arange(0, 210 * 60, 30)
    list_l = [25, 50, 100, 150]

    import matplotlib.pyplot as plt

    plt.style.use("seaborn-paper")
    fig, ax = plt.subplots(figsize=(3.94, 2.76))
    ax.set_xlabel("Time [minute]")
    ax.set_ylabel(u"Temperature [$℃$]")

    for length in list_l:
        temperature = fire(
            t=time,
            fire_load_density_MJm2=600,
            fire_hrr_density_MWm2=0.25,
            room_length_m=length,
            room_width_m=16,
            fire_spread_rate_ms=0.012,
            beam_location_height_m=3,
            beam_location_length_m=length / 2,
            fire_nft_limit_c=1050,
        )

        ax.plot(time / 60, temperature, label="Room length {:4.0f} m".format(length))

    ax.legend(loc=4).set_visible(True)
    ax.set_xlim((-10, 190))
    ax.grid(color="k", linestyle="--")
    plt.tight_layout()
    # plt.show()


def _test_fire_multiple_beam_location():
    time = np.arange(0, 210 * 60, 30)
    length = 100

    import matplotlib.pyplot as plt

    plt.style.use("seaborn-paper")
    fig, ax = plt.subplots(figsize=(3.94, 2.76))
    ax.set_xlabel("Time [minute]")
    ax.set_ylabel("Temperature [$℃$]")

    temperature_list = fire(
        t=time,
        fire_load_density_MJm2=600,
        fire_hrr_density_MWm2=0.25,
        room_length_m=length,
        room_width_m=16,
        fire_spread_rate_ms=0.012,
        beam_location_height_m=3,
        beam_location_length_m=np.linspace(0, length, 12)[1:-1],
        fire_nft_limit_c=1050,
    )

    for temperature in temperature_list:
        ax.plot(time / 60, temperature, label="Room length {:4.0f} m".format(length))

    ax.legend(loc=4).set_visible(True)
    ax.set_xlim((-10, 190))
    ax.grid(color="k", linestyle="--")
    plt.tight_layout()
    # plt.show()


if __name__ == "__main__":
    _test_fire()
    _test_fire_multiple_beam_location()
    # example_plot_interflam()



import json
import os
import types
import warnings
from tkinter import filedialog, Tk, StringVar

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from sfeprapy.func.fire_iso834 import fire as _fire_standard
from sfeprapy.mcs1.mcs1_func_gen import mc_inputs_generator
from sfeprapy.mcs1.mcs1_func_main import main as func_main


class MonteCarloCase:
    # this dictionary object defines all user input variable names as key, and corresponding values:
    # 0 - user input, static;
    # 1 - user input, stochastic; and
    # 2 - derived input
    PARAM = dict(
        window_height=0,
        window_width=0,
        window_open_fraction=1,
        room_breadth=0,
        room_depth=0,
        room_height=0,
        room_wall_thermal_inertia=0,
        fire_mode=0,
        fire_time_step=0,
        fire_iso834_time=2,
        fire_iso834_temperature=2,
        fire_tlim=0,
        fire_load_density=1,
        fire_hrr_density_ubound=0,
        fire_hrr_density_lbound=0,
        fire_hrr_density_mean=0,
        fire_hrr_density_std=0,
        fire_spread_speed=1,
        fire_nft_ubound=1,
        fire_duration=0,
        fire_t_alpha=0,
        fire_gamma_fi_q=0,
        beam_position=1,
        beam_rho=0,
        beam_c=2,
        beam_cross_section_area=0,
        beam_loc_z=0,
        protection_k=0,
        protection_rho=0,
        protection_c=0,
        protection_protected_perimeter=0,
        solver_temperature_target=0,
        solver_fire_duration=0,
        solver_thickness_lbound=0,
        solver_thickness_ubound=0,
        solver_tolerance=0,
        solver_iteration_limit=0,
        index=2,
    )

    def __init__(
        self,
        path_wd: str,
        name: str,
        func: (types.FunctionType, types.BuiltinFunctionType),
    ):
        self._path_wd = path_wd
        self._name = name
        self._func = func

        try:
            os.mkdir(path_wd)
        except:
            pass

        self._input_param = None

        # flags
        self._is_live = False
        self._flag_output_numeric_results = False
        self._flag_output_numeric_fires = False
        self._flag_output_plot_teq = False
        self._flag_output_plot_dist = False

        # user preferences
        self._n_simulations = 1000
        self._n_threads = 2

        self._func_gen = None
        self._func_mcs = None

        # generated data
        self._status = 0
        self._mc_param = None
        self._mc_results = None

        self.timer_mc = None

    @property
    def is_live(self):
        return self._is_live

    @property
    def input_param(self):
        return self._input_param

    @property
    def mc_param(self):
        return self._mc_param

    @property
    def name(self):
        return self._name

    @property
    def mc_param_list(self):
        return self._mc_param.to_dict("index")

    @property
    def n_threads(self):
        return self._n_threads

    @property
    def n_simulations(self):
        return int(self._n_simulations)

    @property
    def mc_results(self):
        return self._mc_results

    @is_live.setter
    def is_live(self, val: bool):
        self._is_live = val

    @input_param.setter
    def input_param(self, val: dict):
        if not isinstance(val, dict):
            raise TypeError(
                "Input parameters variable (input_param) should be a dict object."
            )
        for key, val_ in self.PARAM.items():
            if val_ == 0:
                if key not in val:
                    raise ValueError("Could not find {} in input_param.".format(key))

        self.n_threads = val["n_threads"]
        self.n_simulations = val["n_simulations"]
        self._input_param = val

    @mc_param.setter
    def mc_param(self, val: pd.DataFrame):
        if not isinstance(self, pd.DataFrame):
            raise TypeError(
                "MC parameters variable (mc_param) should be a DataFrame object."
            )
        if not len(val.index) == self.n_simulations:
            raise ValueError(
                "Length mismatch, len(mc_param)={} != len(n_simulations)={}".format(
                    len(val.index), self.n_simulations
                )
            )

        self._mc_param = val

    @mc_results.setter
    def mc_results(self, val: pd.DataFrame):
        if not isinstance(val, pd.DataFrame):
            raise TypeError(
                "MC results variable (mc_results) should be a DataFrame object."
            )
        self._mc_results = val

    @n_threads.setter
    def n_threads(self, val: int):
        if val is None:
            val = 1
        if val < 1:
            warnings.warn(
                "Number of threads can only be a positive integer, your input value is {} and this is changed"
                " to 1.".format(val)
            )
            val = 1
        elif val != int(val):
            warnings.warn(
                "Number of threads can only be a positive integer, your input value is {} and this is changed"
                " to {}.".format(val, int(val))
            )

        self._n_threads = int(val)

    @n_simulations.setter
    def n_simulations(self, val: int):

        if val is None:
            val = int(10)
        if val < 1:
            warnings.warn(
                "Number of threads can only be a positive integer, your input value is {} and this is changed"
                " to 1.".format(val)
            )
            val = int(1)
        elif val != int(val):
            warnings.warn(
                "Number of threads can only be a positive integer, your input value is {} and this is changed"
                " to {}.".format(val, int(val))
            )
            val = int(val)

        self._n_simulations = val

    def in2sc(self):
        # User input to stochastic randomised parameters

        if self.input_param is None:
            raise ValueError("input_param is not defined.")

        input_param = self.input_param

        self.n_simulations = input_param["n_simulations"]
        self.n_threads = input_param["n_threads"]

        # finalise and save to *_in.csv
        fire_time = np.arange(
            0,
            input_param["fire_time_duration"] + input_param["fire_time_step"],
            input_param["fire_time_step"],
        )
        iso834_time = np.arange(0, 6 * 60 * 60, 30)
        iso834_temperature = _fire_standard(iso834_time, 273.15 + 20)

        dict_input_param = self.input_param
        df_mc_params = mc_inputs_generator(**dict_input_param)

        for key, val in self.PARAM.items():
            if val == 0:
                df_mc_params[key] = np.full(self.n_simulations, self.input_param[key])

        df_mc_params["fire_iso834_time"] = [iso834_time] * self.n_simulations
        df_mc_params["fire_iso834_temperature"] = [
            iso834_temperature
        ] * self.n_simulations
        df_mc_params["fire_time"] = [fire_time] * self.n_simulations
        df_mc_params["index"] = np.arange(0, self.n_simulations, 1)

        df_mc_params.index.name = [
            "index"
        ]  # todo is this correct? shouldn't it be a str?

        self._mc_param = df_mc_params

    def mc_sim(self):

        if self.input_param is None:
            raise ValueError("input_param is not defined")
        if self.mc_param is None:
            raise ValueError("mc_param is not defined.")

        if self._is_live:
            df_output = self._func(self._mc_param)
        elif os.path.exists(os.path.join(self._path_wd, self._name + ".csv")):
            # df_output = pd.DataFrame.from_csv(os.path.join(self._path_wd, self._name + '.csv'))
            df_output = pd.read_csv(os.path.join(self._path_wd, self.name + ".csv"))
        else:
            df_output = None

        self._mc_results = df_output

    def res2csv(self, path=None):

        if self.mc_results is None:
            warnings.warn("No results to save.")
            return -1

        if path is None:
            path = os.path.join(self._path_wd, self._name + ".csv")

        name_suffix = 0
        while True:
            try:
                if name_suffix == 0:
                    self.mc_results.to_csv(path)
                    break
                else:
                    self.mc_results.to_csv(path + str(int(name_suffix)))
                    break
            except PermissionError:
                name_suffix += 1

    def res2cdf(self, name: str, cdf_loc: (list, np.ndarray)):
        if self.mc_results is None:
            warnings.warn("No results are extracted.")
            return -1

        x = self.mc_results[name].values
        x = np.sort(x)
        x = np.insert(x, 0, 0)
        y = np.linspace(0, 1, len(x))
        func = interp1d(y, x)

        return func(cdf_loc)

    def mc_results_get(self, name: str):
        return self.mc_results[name].values


class MonteCarlo:
    DEFAULT_TEMP_NAME = "temp"
    DEFAULT_CONFIG_FILE_NAME = "config.json"
    DEFAULT_FIGURE_NAME_TEQ_COMBINED = "teq_combined.png"
    DEFAULT_CONFIG = dict(
        n_threads=1,
        reliability_target=0.8,
        output_fires=0,
        plot_teq_xlim=180,
        plot_style="seaborn-paper",
        plot_figuresize=[3.94, 2.76],
    )

    def __init__(self):
        self._path_master_csv = None

        self._dict_input_param = dict()
        self._dict_input_wd = dict()
        self._dict_input_case_name = dict()

        self._dict_config = dict()

        self._monte_carlo_cases = list()

    @property
    def path_master_csv(self):
        return self._path_master_csv

    @property
    def dict_input_param(self):
        return self._dict_input_param

    @property
    def dict_input_wd(self):
        return self._dict_input_wd

    @property
    def dict_input_case_name(self):
        return self._dict_input_case_name

    @property
    def config(self, key=None):
        if key is None:
            return self._dict_config
        else:
            try:
                return self._dict_config[key]
            except KeyError:
                return None

    @property
    def monte_carlo_cases(self):
        return self._monte_carlo_cases

    @path_master_csv.setter
    def path_master_csv(self, val: str):
        val = os.path.realpath(val)
        if not os.path.isfile(val):
            raise FileNotFoundError(
                "file {path_master_csv} not found.".format(path_master_csv=val)
            )
        self._path_master_csv = val

    @config.setter
    def config(self, val: dict):
        if not isinstance(val, dict):
            raise TypeError("config should be a dict object.")
        self._dict_config = val

    @monte_carlo_cases.setter
    def monte_carlo_cases(self, val: MonteCarloCase):
        if not isinstance(self._monte_carlo_cases, list):
            self._monte_carlo_cases = [val]
        else:
            flag_exist = False
            for i, mc_case in enumerate(self._monte_carlo_cases):
                if mc_case.name == val.name:
                    self._monte_carlo_cases[i] = val
                    flag_exist = True

            if not flag_exist:
                self._monte_carlo_cases.append(val)

    def select_input_file(self, path_master_csv: str = None):

        if path_master_csv is None:
            self.path_master_csv = self._gui_file_path()
        else:
            self.path_master_csv = os.path.realpath(
                path_master_csv.replace('"', "").replace("'", "")
            )

        path_config = os.path.join(
            os.path.dirname(self.path_master_csv), self.DEFAULT_CONFIG_FILE_NAME
        )

        if os.path.isfile(path_config):
            with open(path_config, "r") as f:
                self.config = json.load(f)
        else:
            warnings.warn("config.json not found, default config parameters are used.")
            self.config = self.DEFAULT_CONFIG

    def make_input_param(self):
        if self.path_master_csv is None:
            raise ValueError("csv file path (path_master_csv) not defined.")
        if self.config is None:
            raise ValueError("config is not defined.")

        path_input_file_csv = self.path_master_csv
        config = self.config

        # csv to dict
        if path_input_file_csv.endswith(".csv"):
            df_input_params = pd.read_csv(path_input_file_csv).set_index("PARAMETERS")
        elif path_input_file_csv.endswith(".xlsx"):
            df_input_params = pd.read_excel(path_input_file_csv).set_index("PARAMETERS")
        else:
            df_input_params = []

        df_input_params = df_input_params.append(
            pd.Series(
                {k: config["n_threads"] for k in df_input_params.columns.values},
                name="n_threads",
            )
        )

        dict_dict_input_params = df_input_params.to_dict()

        # dict to json (list)
        for case_name, input_param in dict_dict_input_params.items():
            MCC = MonteCarloCase(
                path_wd=os.path.join(
                    os.path.dirname(path_input_file_csv), self.DEFAULT_TEMP_NAME
                ),
                name=case_name,
                func=func_main,
            )
            MCC.input_param = input_param
            MCC.is_live = int(input_param["is_live"]) == 1
            MCC.in2sc()

            self.monte_carlo_cases = MCC

    def make_mc_params(self):
        if self.monte_carlo_cases is None:
            raise ValueError("no monte carlo case has been set up.")

        for case in self.monte_carlo_cases:
            case.in2sc()

    def run_mc(self):

        for case in self.monte_carlo_cases:
            case.mc_sim()
            case.res2csv()

    def out_combined_ky(self):

        dict_res = dict()

        for case in self.monte_carlo_cases:
            if case.is_live:
                dict_res[case.name] = case.mc_results_get("strength_reduction_factor")

        p = os.path.join(os.path.dirname(self.path_master_csv), "ky.csv")
        df_ = pd.DataFrame.from_dict(dict_res)
        df_.to_csv(p)

    def out_combined_T(self):

        dict_res = dict()

        for case in self.monte_carlo_cases:
            if case.is_live:
                dict_res[case.name] = case.mc_results_get(
                    "solver_steel_temperature_solved"
                )

        p = os.path.join(os.path.dirname(self.path_master_csv), "T.csv")
        df_ = pd.DataFrame.from_dict(dict_res)
        df_.to_csv(p)

    def _plot_teq(self):
        plot_data = {}
        for case in self.monte_carlo_cases:
            plot_data[case.name] = dict(
                x=case.mc_results_get(""), y=np.linspace(0, 1, case.n_simulations)
            )

        self._plot_figure(
            path_save_figure=os.path.join(
                os.path.dirname(self.path_master_csv),
                self.DEFAULT_FIGURE_NAME_TEQ_COMBINED,
            ),
            data=plot_data,
            plot_xlim=(0, 1),
        )

    def _get_results_teq(self):

        for case in self.monte_carlo_cases:
            case.mc_results_get("solver_steel_temperature_solved")

    @staticmethod
    def _plot_figure(
        path_save_figure: str,
        data: dict,
        plot_xlim: tuple,
        plot_figuresize=(3.94, 2.76),
    ):

        teq_fig, teq_ax = plt.subplots(figsize=plot_figuresize)
        teq_ax.set_xlim([0, plot_xlim])

        for case_name, plot_data in data.items():
            teq_ax.plot(plot_data["x"], plot_data["y"], label=case_name, linewidth=1)
            teq_ax.set_ylabel("y-axis label")
            teq_ax.set_xlabel("x-axis label")
            teq_ax.set_xticks(ticks=np.arange(0, plot_xlim + 0.001, 30))

        teq_ax.legend().set_visible(True)
        teq_ax.legend(prop={"size": 7})
        plt.tight_layout()
        plt.savefig(path_save_figure, transparent=True, bbox_inches="tight", dpi=300)

        plt.clf()
        plt.close()

    @staticmethod
    def _gui_file_path():
        # get a list of dict()s representing different scenarios
        root = Tk()
        root.withdraw()
        folder_path = StringVar()

        path_input_file_csv = filedialog.askopenfile(
            title="Select Input File",
            parent=root,
            filetypes=[("SPREADSHEET", [".csv", ".xlsx"])],
            mode="r",
        )
        folder_path.set(path_input_file_csv)
        root.update()

        try:
            path_input_file_csv = os.path.realpath(path_input_file_csv.name)
            return path_input_file_csv
        except AttributeError:
            raise FileNotFoundError("file not found.")


if __name__ == "__main__":
    MC = MonteCarlo()
    MC.select_input_file()
    MC.make_input_param()
    MC.make_mc_params()
    MC.run_mc()



import numpy as np


def fire(
    t, A_t, A_f, A_v, h_eq, q_fd, lambda_, rho, c, t_lim, temperature_initial=293.15
):
    """Function Description: (SI UNITS ONLY)
    This function calculates the time-temperature curve according to Eurocode 1 part 1-2, Appendix A.
    :param t: numpy.ndarray, [s], time evolution.
    :param A_t: float, [m2], total surface area (including openings).
    :param A_f: float, [m2], floor area.
    :param A_v: float, [m2], opening area.
    :param h_eq: float, [m2], opening height.
    :param q_fd: float, [J/m2], fuel density.
    :param lambda_: float, [K/kg/m], lining thermal conductivity.
    :param rho: float, [kg/m3], lining density.
    :param c: float, [J/K/kg], lining thermal capacity.
    :param t_lim: float, [s], limiting time for the fire.
    :return T_g: numpy.ndarray, [s], temperature evolution.
    """
    # Reference: Eurocode 1991-1-2; Jean-Marc Franssen, Paulo Vila Real (2010) - Fire Design of Steel Structures

    # UNITS: SI -> Equations
    q_fd /= 1e6  # [J/m2] -> [MJ/m2]
    t_lim /= 3600  # [s] -> [hr]
    t = t / 3600  # [s] -> [hr]
    temperature_initial -= 273.15  # [K] -> [C]

    # ACQUIRING REQUIRED VARIABLES
    # t = np.arange(time_start, time_end, time_step, dtype=float)

    b = (lambda_ * rho * c) ** 0.5
    O = A_v * h_eq ** 0.5 / A_t
    q_td = q_fd * A_f / A_t
    Gamma = ((O / 0.04) / (b / 1160)) ** 2

    t_max = 0.0002 * q_td / O

    # check criteria
    # todo: raise warnings to higher level
    # if not 50 <= q_td <= 1000:
    #     if is_cap_q_td:
    #         msg = "q_td ({:4.1f}) EXCEEDED [50, 1000] AND IS CAPPED.".format(q_td, is_cap_q_td)
    #     else:
    #         msg = "q_td ({:4.1f}) EXCEEDED [50, 1000] AND IS UNCAPPED.".format(q_td)
    #     warnings.warn(msg)

    # CALCULATION
    def _temperature_heating(t_star, temperature_initial):
        # eq. 3.12
        T_g = 1325 * (
            1
            - 0.324 * np.exp(-0.2 * t_star)
            - 0.204 * np.exp(-1.7 * t_star)
            - 0.472 * np.exp(-19 * t_star)
        )
        T_g += temperature_initial
        return T_g

    def _temperature_cooling_vent(t_star_max, T_max, t_star):  # ventilation controlled
        # eq. 3.16
        if t_star_max <= 0.5:
            T_g = T_max - 625 * (t_star - t_star_max)
        elif 0.5 < t_star_max < 2.0:
            T_g = T_max - 250 * (3 - t_star_max) * (t_star - t_star_max)
        elif 2.0 <= t_star_max:
            T_g = T_max - 250 * (t_star - t_star_max)
        else:
            T_g = np.nan
        return T_g

    def _temperature_cooling_fuel(
        t_star_max, T_max, t_star, Gamma, t_lim
    ):  # fuel controlled
        # eq. 3.22
        if t_star_max <= 0.5:
            T_g = T_max - 625 * (t_star - Gamma * t_lim)
        elif 0.5 < t_star_max < 2.0:
            T_g = T_max - 250 * (3 - t_star_max) * (t_star - Gamma * t_lim)
        elif 2.0 <= t_star_max:
            T_g = T_max - 250 * (t_star - Gamma * t_lim)
        else:
            T_g = np.nan
        return T_g

    def _variables(t, Gamma, t_max):
        t_star = Gamma * t
        t_star_max = Gamma * t_max
        return t_star, t_star_max

    def _variables_2(t, t_lim, q_td, b, O):
        O_lim = 0.0001 * q_td / t_lim
        Gamma_lim = ((O_lim / 0.04) / (b / 1160)) ** 2

        if O > 0.04 and q_td < 75 and b < 1160:
            k = 1 + ((O - 0.04) / (0.04)) * ((q_td - 75) / (75)) * ((1160 - b) / (1160))
            Gamma_lim *= k

        t_star_ = Gamma_lim * t
        t_star_max_ = Gamma_lim * t_lim
        return t_star_, t_star_max_

    t_star, t_star_max = _variables(t, Gamma, t_max)

    if t_max >= t_lim:  # ventilation controlled fire
        T_max = _temperature_heating(t_star_max, temperature_initial)
        T_heating_g = _temperature_heating(Gamma * t, temperature_initial)
        T_cooling_g = _temperature_cooling_vent(t_star_max, T_max, t_star)
        fire_type = "ventilation controlled"
    else:  # fuel controlled fire
        t_star_, t_star_max_ = _variables_2(t, t_lim, q_td, b, O)
        T_max = _temperature_heating(t_star_max_, temperature_initial)
        T_heating_g = _temperature_heating(t_star_, temperature_initial)
        T_cooling_g = _temperature_cooling_fuel(t_star_max, T_max, t_star, Gamma, t_lim)
        fire_type = "fuel controlled"

    T_g = np.minimum(T_heating_g, T_cooling_g)
    T_g[T_g < temperature_initial] = temperature_initial

    # UNITS: Eq. -> SI
    t *= 3600
    T_g += 273.15

    return T_g


def example_plot_interflam():
    import matplotlib.pyplot as plt
    import seaborn as sns

    sns.set_style("ticks")
    sns.set_context(context="paper")

    fig, ax = plt.subplots(figsize=(3.94, 2.76))
    ax.set_xlabel("Time [minute]")
    ax.set_ylabel(r"Temperature [$^{\circ}C$]")

    # define geometry
    w = 16
    l = 32
    h = 3

    h_v = 2

    oo = (0.02, 0.04, 0.06, 0.10, 0.14, 0.20)  # desired opening factor

    def opening_factor_2_window_width(of, w, l, h, h_v):
        # opening factor = Av * sqrt(hv) / At
        # Av = h_v * w_v
        # of * At / sqrt(h_v) = h_v * w_v
        # w_v = of * At / sqrt(h_v) / h_v
        return of * (2 * (w * l + l * h + h * w)) / h_v ** 0.5 / h_v

    w_v_ = [opening_factor_2_window_width(o, w, l, h, h_v) for o in oo]

    print(w_v_)

    # Calculate fire curves
    for w_v in w_v_:
        x = np.arange(0, 5 * 60 * 60, 1)
        y = fire(
            t=x,
            A_v=h_v * w_v,
            A_t=2 * (w * h + h * l + l * w),
            A_f=w * l,
            h_eq=h_v,
            q_fd=600e6,
            lambda_=1,
            rho=1,
            c=720 ** 2,
            t_lim=20 * 60,
            temperature_initial=293.15,
        )
        ax.plot(
            x / 60,
            y - 273.15,
            label="Opening Factor {:.2f}".format(
                (h_v * w_v) * h_v ** 0.5 / (2 * (w * h + h * l + l * w))
            ),
        )

    ax.legend().set_visible(True)
    ax.grid(color="k", linestyle="--")
    ax.set_ylim((0, 1400))
    ax.set_xlim((0, 300))
    plt.tight_layout()

    plt.savefig(fname="fire-ec.png", dpi=300)
    plt.show()


def _test_fire():
    """This is a test function to `fire` within this module `fire_parametric_ec`, it compares the function against
    Figure 7 in Holicky, M. et al [1].
    yan fu, 1 oct 2018

    REFERENCES
    [1] Holicky, M., Meterna, A., Sedlacek, G. and Schleich, J.B., 2005. Implementation of eurocodes, handbook 5, design
    of buildings for the fire situation. Leonardo da Vinci Pilot Project: Luxembourg."""

    import copy
    import pandas

    # LOAD VERIFICATION DATA
    # data are from the referenced document
    from io import StringIO

    verification_data = StringIO(
        """time_1,temperature_1,time_2,temperature_2,time_3,temperature_3,time_4,temperature_4,time_5,temperature_5,time_6,temperature_6,time_7,temperature_7
        20,689.1109391,20,689.1109391,20,689.1109391,20,843.9616854,20,741.3497451,20,601.424372,20,269.3348197
        30,782.7282474,30,782.7317977,30,782.7317977,30,946.9144525,30,827.5114674,30,726.7651987,30,430.1234077
        40,167.4009827,40,402.4756096,40,508.8188932,40,781.2034026,40,885.6845648,40,777.4756096,40,533.0726245
        50,20,50,20,50,236.7716603,50,540.8761379,50,818.8612125,50,813.2641976,50,604.3089737
        60,20,60,20,60,20,60,304.2731159,60,692.3328174,60,843.4522203,60,651.2880412
        70,20,70,20,70,20,70,69.53931579,70,565.8079725,70,780.3602113,70,685.2109576
        80,20,80,20,80,20,80,20,80,441.145249,80,719.1303236,80,709.8019654"""
    )
    verification_data = pandas.read_csv(
        verification_data, skip_blank_lines=True, skipinitialspace=True
    )

    # CALCULATE TIME TEMPERATURE CURVE BASED ON THE VERIFICATION DATA INPUTS

    # prepare inputs
    kws = dict(
        A_t=360,
        A_f=100,
        h_eq=1,
        q_fd=600e6,
        lambda_=1,
        rho=1,
        c=2250000,
        t_lim=20 * 60,
        t=np.arange(0, 2 * 60 * 60 + 1, 1),
        temperature_initial=293.15,
    )

    # define opening area
    A_v_list = [72, 50.4, 36.000000001, 32.4, 21.6, 14.4, 7.2]

    # calculate fire curves
    x1_list = []  # calculated time array
    y1_list = []  # calculated temperature array
    for i in A_v_list:
        y = fire(A_v=i, **copy.copy(kws))
        x = np.arange(0, 2 * 60 * 60 + 1, 1) + 10 * 60
        x1_list.append(x / 60)
        y1_list.append(y - 273.15)

    # COMPARE CALCULATED AND THE VERIFICATION DATA

    def r_square(x1, y1, x2, y2):
        from scipy.interpolate import interp1d
        from scipy import stats

        slope, intercept, r_value, p_value, std_err = stats.linregress(
            y2, interp1d(x1, y1)(x2)
        )
        return r_value ** 2

    for i in range(len(A_v_list)):
        x1 = x1_list[i]
        y1 = y1_list[i]
        x2 = verification_data["time_{}".format(int(i + 1))]
        y2 = verification_data["temperature_{}".format(int(i + 1))]
        assert r_square(x1, y1, x2, y2) > 0.99


if __name__ == "__main__":
    # exmaple_plot_interflam()
    _test_fire()



# -*- coding: utf-8 -*-


from sfeprapy.func.fire_travelling import _test_fire as test_fire_travelling
from sfeprapy.func.fire_travelling import (
    _test_fire_backup as test_fire_travelling_backup,
)
from sfeprapy.func.fire_travelling import (
    _test_fire_multiple_beam_location as test_fire_travelling_multiple,
)

test_fire_travelling()
test_fire_travelling_backup()
test_fire_travelling_multiple()



import numpy as np


if __name__ == "__main__":

    L = 1  # length of the medium
    Nx = 10  # number of nodes (layers) in the medium
    T = 10  # upper time limit
    Nt = 100  # number of nodes in the time limit
    a = 500  # conductivity coefficient

    x = np.linspace(0, L, Nx + 1)  # mesh points in space
    dx = x[1] - x[0]
    t = linspace(0, T, Nt + 1)  # mesh points in time
    dt = t[1] - t[0]
    F = a * dt / dx ** 2
    u = zeros(Nx + 1)  # unknown u at new time level
    u_1 = zeros(Nx + 1)  # u at the previous time level

    # Set initial condition u(x,0) = I(x)
    for i in range(0, Nx + 1):
        u_1[i] = I(x[i])

    for n in range(0, Nt):
        # Compute u at inner mesh points
        for i in range(1, Nx):
            u[i] = u_1[i] + F * (u_1[i - 1] - 2 * u_1[i] + u_1[i + 1])

        # Insert boundary conditions
        u[0] = 0
        u[Nx] = 0

        # Update u_1 before next step
        u_1[:] = u



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from io import StringIO
import warnings


def steel_specific_heat_carbon_steel(temperature):
    """
    :param temperature: {float} [K]
    :return: {float} [J/kg/K]
    """
    temperature -= 273.15
    if 20 <= temperature < 600:
        return (
            425
            + 0.773 * temperature
            - 1.69e-3 * np.power(temperature, 2)
            + 2.22e-6 * np.power(temperature, 3)
        )
    elif 600 <= temperature < 735:
        return 666 + 13002 / (738 - temperature)
    elif 735 <= temperature < 900:
        return 545 + 17820 / (temperature - 731)
    elif 900 <= temperature <= 1200:
        return 650
    else:
        return 0


if __name__ == "__main__":

    import timeit

    test_number = 100000

    def aa():
        steel_specific_heat_carbon_steel(np.random.rand() * 1180 + 20 + 273.15)

    print(timeit.timeit(aa, number=test_number))
    print(timeit.timeit(bb, number=test_number))



# -*- coding: utf-8 -*-
import numpy as np


def unprotected_steel_eurocode(
    time,
    temperature_ambient,
    perimeter_section,
    area_section,
    perimeter_box,
    density_steel,
    c_steel_T,
    h_conv,
    emissivity_resultant,
):
    """
    SI UNITS FOR INPUTS AND OUTPUTS.
    :param time:
    :param temperature_ambient:
    :param perimeter_section:
    :param area_section:
    :param perimeter_box:
    :param density_steel:
    :param c_steel_T:
    :param h_conv:
    :param emissivity_resultant:
    :return:
    """

    # Create steel temperature change array s
    temperature_rate_steel = time * 0.0
    temperature_steel = time * 0.0
    heat_flux_net = time * 0.0
    c_s = time * 0.0

    k_sh = (
        0.9 * (perimeter_box / area_section) / (perimeter_section / area_section)
    )  # BS EN 1993-1-2:2005 (e4.26a)
    F = perimeter_section
    V = area_section
    rho_s = density_steel
    h_c = h_conv
    sigma = 56.7e-9  # todo: what is this?
    epsilon = emissivity_resultant

    time_, temperature_steel[0], c_s[0] = iter(time), temperature_ambient[0], 0.0
    next(time_)
    for i, v in enumerate(time_):
        i += 1
        T_f, T_s_ = (
            temperature_ambient[i],
            temperature_steel[i - 1],
        )  # todo: steel specific heat
        c_s[i] = c_steel_T(temperature_steel[i - 1] + 273.15)

        # BS EN 1993-1-2:2005 (e4.25)
        a = h_c * (T_f - T_s_)
        b = sigma * epsilon * (np.power(T_f, 4) - np.power(T_s_, 4))
        c = k_sh * F / V / rho_s / c_s[i]
        d = time[i] - time[i - 1]

        heat_flux_net[i] = a + b

        temperature_rate_steel[i] = c * (a + b) * d
        temperature_steel[i] = temperature_steel[i - 1] + temperature_rate_steel[i]

    return temperature_steel, temperature_rate_steel, heat_flux_net, c_s



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
import os


def thermal(property_name):
    dir_package = os.path.dirname(os.path.abspath(__file__))

    dir_files = {
        "thermal conductivity (thomas2002)": "k_1_T_intumescent_thomas2002.csv"
    }

    # read file
    data = pd.read_csv(
        "/".join([dir_package, dir_files[property_name]]), delimiter=",", dtype=float
    )
    x, y = data.values[:, 0], data.values[:, 1]

    return interp1d(x, y)


# if __name__ == "__main__":
#     temperature = np.arange(20.,1200.+0.5,0.5) + 273.15
#     prop = temperature * 0
#     for i,v in enumerate(temperature):
#         prop[i] = steel_specific_heat_carbon_steel(v)
#     df = pd.DataFrame(
#         {
#             "Temperature [K]": temperature,
#             "Specific Heat Capacity [J/kg/K]": prop
#         }
#     )



"""SfePrapy CLI Help.
Usage:
    sfeprapy mcs0 [-p=<int>] <file_name>
    sfeprapy mcs0 template <file_name>
    sfeprapy mcs0 figure <file_name>
    sfeprapy distfit [--data_t=<int>] [--dist_g=<int>] <file_name>

Options:
    --data_t=<int>  an integer indicating data type:
                    0   (default) samples only, a single column data.
                    1   PDF, two columns data containing x, y, without heading.
                    2   CDF, two columns data containing x, y, without heading.
    --dist_g=<int>  an integer indicating what distribution group to be used for fitting the data:
                    0   fit to all available distributions
                    1   (default) fit to common distribution types
    -p=<int>        to define number of processes for MCS, positive integer only.
    -h --help       to show this message.

Commands:
    mcs0            Monte Carlo Simulation to solve equivalent time exposure in ISO 834 fire, method 0.
    mcs0 template   save example input file to <file_name>.
    mcs0 figure     produce figure from the output file <file_name>.
"""

TEST_MODE = False

from docopt import docopt


def main():
    import os
    import sfeprapy

    # sfeprapy.check_pip_upgrade()

    def _test(outcome: bool = True):
        if TEST_MODE:
            return outcome

    arguments = docopt(__doc__)

    arguments["<file_name>"] = os.path.realpath(arguments["<file_name>"])

    if arguments["mcs0"]:
        _test()

        if arguments["figure"]:
            _test()
            from sfeprapy.mcs0.__main__ import save_figure as mcs0_figure

            mcs0_figure(fp_mcs0_out=arguments["<file_name>"])

        if arguments["template"]:
            _test()
            from sfeprapy.mcs0 import EXAMPLE_INPUT_CSV

            with open(arguments["<file_name>"], "w+") as f:
                f.write(EXAMPLE_INPUT_CSV)

        if not (arguments["figure"] or arguments["template"]):
            _test()
            from sfeprapy.mcs0.__main__ import main as mcs0

            fp_mcs_in = arguments["<file_name>"]
            n_threads = arguments["-p"] or 2
            mcs0(fp_mcs_in=fp_mcs_in, n_threads=int(n_threads))

    elif arguments["distfit"]:
        _test()

        # Prerequisites
        from sfeprapy.func.stats_dist_fit import auto_fit_2

        # Default values
        data_type = arguments["--data_t"] or 2
        distribution_list = arguments["--dist_g"] or 1

        # Main
        auto_fit_2(
            data_type=int(data_type),
            distribution_list=int(distribution_list),
            data=arguments["<file_name>"],
        )
    else:
        _test(False)
        print("instruction unclear.")


def _test_main():
    global TEST_MODE
    TEST_MODE = True
    assert docopt(__doc__, ["mcs0", "file_name"])
    assert docopt(__doc__, ["mcs0", "template", "file_name"])
    assert docopt(__doc__, ["mcs0", "figure", "file_name"])
    assert docopt(__doc__, ["distfit", "file_name"])



# -*- coding: utf-8 -*-

from sfeprapy.func.mcs_gen import _test_dict_flatten as test_mcs_gen_dict_flatten
from sfeprapy.func.mcs_gen import (
    _test_random_variable_generator as test_mcs_gen_random_variable_generator,
)

test_mcs_gen_dict_flatten()
test_mcs_gen_random_variable_generator()



# Sub-functions for 1D heat transfer code
# __main__ will main_args a non-linear 1D heat transfer analysis
# Danny Hopkin
# OFR Consultants
# 15/05/2019

#  Conversion from DegC to DegK

import numpy as np
from matplotlib import pyplot as plt


def ISO834_ft(t):

    #  returns the ISO curve where t is [s]
    tmin = t / 60
    tiso = 345 * np.log10((8 * tmin) + 1) + 20
    return tiso


def Common_CtoK(T):

    # Returns Kelvin

    TK = T + 273
    return TK


#  Calculation of flux abosrbed by first element


def ONEDHT_QINC(TGAS, TSURF, EMIS, hc):

    # Get absorbed heat flux (in Watts)

    QRAD = (
        0.0000000567 * EMIS * ((Common_CtoK(TGAS) ** 4) - (Common_CtoK(TSURF) ** 4))
    )  # Radiation component
    QCONV = hc * (Common_CtoK(TGAS) - Common_CtoK(TSURF))  # Convective component
    QINC = QRAD + QCONV  # Sum components and return value
    return QINC


# Calculation of heat flux leaving unexposed element


def ONEDHT_QOUT(TSURF, TAMB, EMIS, hc):

    # Heat flux lost to ambient on unexposed face (in Watts)

    QRAD = (
        0.0000000567 * EMIS * ((Common_CtoK(TSURF) ** 4) - (Common_CtoK(TAMB) ** 4))
    )  # Radiation component
    QCONV = hc * (Common_CtoK(TSURF) - Common_CtoK(TAMB))  # Convective component
    QOUT = QRAD + QCONV  # Sum components and return value
    return QOUT


# Calculation of temperature of first element


def ONEDHT_ELEM1(Qinc, T1, T2, LAMDA1, LAMDA2, dx, dt, Cp, Rho):

    # Calculate temperature of first element given incoming flux Q1

    a1 = (2 * dt) / (Rho * Cp * dx)  # calculate diffusivity
    a2 = (LAMDA1 + LAMDA2) / 2  # calculate mean conductivity
    a3 = (T1 - T2) / dx  # Calculate temperature gradient

    dT1 = a1 * (Qinc - (a2 * a3))  #  Calculate change in temperature in time step

    T1new = max(T1 + dT1, 20)  # Calculate new temperature

    return T1new


# Calculation of temperature for intermediate elements


def ONEDHT_ELEMJ(TJN1, TJ, TJP1, LAMDAJN1, LAMDAJ, LAMDAJP1, dx, dt, Cp, Rho):

    # Calculate temperature of element j, using temps for j-1 and j+1
    b1 = dt / (Rho * Cp * (dx ** 2))  # calculate diffusivity
    b2 = (LAMDAJN1 + LAMDAJ) / 2  # calculate mean conductivity1
    b3 = TJN1 - TJ  # Calculate delta T1
    b4 = (LAMDAJ + LAMDAJP1) / 2  # calculate mean conductivity2
    b5 = TJ - TJP1  # Calculate delta T2
    dTJ = b1 * ((b2 * b3) - (b4 * b5))  # Calculate change in temperature in time step
    Tjnew = TJ + dTJ  # Calculate new temperature
    return Tjnew


# Calculation of temperature of final element


def ONEDHT_ELEMF(Qout, TFN1, TF, LAMDAFN1, LAMDAF, dx, dt, Cp, Rho):

    # Calculate temperature of the final element

    c1 = (2 * dt) / (Rho * Cp * dx)  # calculate diffusivity
    c2 = (LAMDAFN1 + LAMDAF) / 2  # calculate mean conductivity
    c3 = (TFN1 - TF) / dx  # Calculate temperature gradient

    dTF = c1 * ((c2 * c3) - Qout)  # calculate change in temperature in time step
    Tfnew = TF + dTF  # calculate new temperature
    return Tfnew


import warnings


def c_steel_T(temperature):

    if temperature < 20:
        warnings.warn("Temperature ({:.1f} °C) is below 20 °C".format(temperature))
        return 425 + 0.773 * 20 - 1.69e-3 * np.power(20, 2) + 2.22e-6 * np.power(20, 3)
    if 20 <= temperature < 600:
        return (
            425
            + 0.773 * temperature
            - 1.69e-3 * np.power(temperature, 2)
            + 2.22e-6 * np.power(temperature, 3)
        )
    elif 600 <= temperature < 735:
        return 666 + 13002 / (738 - temperature)
    elif 735 <= temperature < 900:
        return 545 + 17820 / (temperature - 731)
    elif 900 <= temperature <= 1200:
        return 650
    elif temperature > 1200:
        warnings.warn(
            "Temperature ({:.1f} °C) is greater than 1200 °C".format(temperature)
        )
        return 650
    else:
        warnings.warn("Temperature ({:.1f} °C) is outside bound.".format(temperature))
        return 0


def k_steel_T(temperature):
    if temperature < 20:
        warnings.warn("Temperature ({:.1f} °C) is below 20 °C".format(temperature))
        return 54 - 3.33e-2 * 20
    if temperature < 800:
        return 54 - 3.33e-2 * temperature
    elif temperature <= 1200:
        return 27.3
    elif temperature > 1200:
        warnings.warn(
            "Temperature ({:.1f} °C) is greater than 1200 °C".format(temperature)
        )
        return 27.3
    else:
        warnings.warn("Temperature ({:.1f} °C) is outside bound.".format(temperature))
        return 0


if __name__ == "__main__":

    import copy
    from scipy.interpolate import interp1d

    # Set geometrical parameters & simulation time frame

    nodes = 100  # set the number of nodes
    dx = 0.1 / nodes  # in m
    temp_arr = (
        np.ones(nodes) * 20
    )  # creates initial temperature nodal array with ambient condition
    endtime = 600  # number of timesteps
    dt = 0.025  # time step increment
    nth = 1000  # Temperature profile will be produced every nth time step
    endtimestep = int(endtime / dt)  # calculate total number of timesteps
    node_array = np.ones(nodes)  # Create depth array for temperature profile

    # Set heat transfer properties based upon initial temperature conditions

    lamda_temp_arr = [0, 1200]  #  Temperature array for conductivity [DegC]
    lamda_input_arr = [40, 40]  #  Corresponding conductivity input [W/m.K]
    # lamda_temp_arr = [19,500,1200]                              #  Temperature array for conductivity [DegC]
    # lamda_input_arr = [0.12,0.11, 1]                             #  Corresponding conductivity input [W/m.K]
    lamda_interp = interp1d(lamda_temp_arr, lamda_input_arr)  #  Interpolation function

    cp_temp_arr = [0, 1200]  #  Temperature array for specific heat [DegC]
    cp_input_arr = [450, 450]  #  Corresponding specific heat input [J/kg.K]
    # cp_temp_arr = [19, 99, 100, 120, 121, 1200]                 #  Temperature array for specific heat [DegC]
    # cp_input_arr =[950, 950, 2000, 2000, 950, 950]              #  Corresponding specific heat input [J/kg.K]
    cp_interp = interp1d(cp_temp_arr, cp_input_arr)  #  Interpolation function

    rho_temp_arr = [0, 1200]  #  Temperature array for density [DegC]
    rho_input_arr = [9850, 9850]
    # rho_temp_arr = [19,1200]                                    #  Temperature array for density [DegC]
    # rho_input_arr =[450,400]                                    #  Corresponding density input [kg/m3]
    rho_interp = interp1d(rho_temp_arr, rho_input_arr)  #  Interpolation function

    # Boundary conditions

    hchot = 25  #  hot side convection coefficient [W/m2.K]
    hccold = 9  # cold side convection coefficient [W/m2.K]
    effemis = 0.8  # net emissivity - applies to all faces [-]
    tamb = 20  # ambient temp on unexposed face [DegC]

    list_temp_arr = list()  # Create storage matrix

    #  Main heat transfer solver

    print("SOLVING")

    for tstep in range(1, endtimestep + 1):

        # Update thermal property array at each time time
        print(temp_arr)

        # lamdaNL = lamda_interp(temp_arr)
        # cpNL = cp_interp(temp_arr)
        # rhoNL = rho_interp(temp_arr)
        lamdaNL = np.array([k_steel_T(temp) for temp in temp_arr])
        cpNL = np.array([c_steel_T(temp) for temp in temp_arr])
        rhoNL = np.full_like(temp_arr, 7850)

        time = dt * tstep  #  current simulation time
        gas_temp = ISO834_ft(time)  #  gas temp in DegC
        gas_temp = 1000

        if tstep / 10 == int(tstep / 10):
            print("Time = ", time, " s ", "Gas temperature = ", gas_temp, " DegC")

        # First element calculations

        Qinc = ONEDHT_QINC(gas_temp, temp_arr[0], effemis, hchot)
        temp_arr[0] = ONEDHT_ELEM1(
            Qinc,
            temp_arr[0],
            temp_arr[1],
            lamdaNL[0],
            lamdaNL[1],
            dx,
            dt,
            cpNL[0],
            rhoNL[0],
        )

        # Intermediate element calculations

        for nodenum in range(1, nodes - 1):
            temp_arr[nodenum] = ONEDHT_ELEMJ(
                temp_arr[nodenum - 1],
                temp_arr[nodenum],
                temp_arr[nodenum + 1],
                lamdaNL[nodenum - 1],
                lamdaNL[nodenum],
                lamdaNL[nodenum + 1],
                dx,
                dt,
                cpNL[nodenum],
                rhoNL[nodenum],
            )

        # Final element calculations

        Qout = ONEDHT_QOUT(temp_arr[nodes - 1], tamb, effemis, hccold)
        temp_arr[nodes - 1] = ONEDHT_ELEMF(
            Qout,
            temp_arr[nodes - 2],
            temp_arr[nodes - 1],
            lamdaNL[nodes - 2],
            lamdaNL[nodes - 1],
            dx,
            dt,
            cpNL[nodes - 1],
            rhoNL[nodes - 1],
        )

        # Store results

        list_temp_arr.append(copy.copy(temp_arr))

    # Generate depth array [mm]

    for x in range(0, nodes):
        node_array[x] = x * dx * 1000

    # Plot temperature profiles at every nth time step

    import seaborn as sns

    sns.set_style("ticks")
    fig, ax = plt.subplots(figsize=(4, 3))
    for i, v in enumerate(list_temp_arr):
        if i / nth == int(i / nth):
            ax.plot(node_array, v)
            ax.grid(True)
            ax.set_xlabel("Depth [mm]")
            ax.set_ylabel("Temperature [$^\circ C$]")

    ax.plot(node_array, list_temp_arr[-1])

    plt.tight_layout()
    plt.show()



# -*- coding: utf-8 -*-

from io import StringIO

import numpy as np
import pandas as pd
import scipy.stats as stats
from scipy.interpolate import interp1d


def gumbel_r_(mean: float, sd: float, **_):

    # parameters Gumbel W&S
    alpha = 1.282 / sd
    u = mean - 0.5772 / alpha

    # parameters Gumbel scipy
    scale = 1 / alpha
    loc = u

    return dict(loc=loc, scale=scale)


def lognorm_(mean: float, sd: float, **_):

    cov = sd / mean

    sigma_ln = np.sqrt(np.log(1 + cov ** 2))
    miu_ln = np.log(mean) - 1 / 2 * sigma_ln ** 2

    s = sigma_ln
    loc = 0
    scale = np.exp(miu_ln)

    return dict(s=s, loc=loc, scale=scale)


def norm_(mean: float, sd: float, **_):

    loc = mean
    scale = sd

    return dict(loc=loc, scale=scale)


def uniform_(ubound: float, lbound: float, **_):

    if lbound > ubound:
        lbound += ubound
        ubound = lbound - ubound
        lbound -= ubound

    loc = lbound
    scale = ubound - lbound

    return dict(loc=loc, scale=scale)


def random_variable_generator(dict_in: dict, num_samples: int):
    """Generates samples of defined distribution. This is build upon scipy.stats library.

    :param dict_in:     distribution inputs, required keys are distribution dependent, should be align with inputs
                        required in the scipy.stats. Additional compulsory keys are:
                            `dist`: str, distribution type;
                            `ubound`: float, upper bound of the sampled values; and
                            `lbound`: float, lower bound of the sampled values.
    :param num_samples: number of samples to be generated.
    :return samples:    sampled values based upon `dist` in the range [`lbound`, `ubound`] with `num_samples` number of
                        values.
    """

    # assign distribution type
    dist_0 = dict_in["dist"]
    dist = dict_in["dist"]

    # assign distribution boundary (for samples)
    ubound = dict_in["ubound"]
    lbound = dict_in["lbound"]

    # sample CDF points (y-axis value)
    def generate_cfd_q(dist_, dist_kw_, lbound_, ubound_):
        cfd_q_ = np.linspace(
            getattr(stats, dist_).cdf(x=lbound_, **dist_kw_),
            getattr(stats, dist_).cdf(x=ubound_, **dist_kw_),
            num_samples,
        )
        samples_ = getattr(stats, dist_).ppf(q=cfd_q_, **dist_kw_)
        return samples_

    # convert human distribution parameters to scipy distribution parameters
    if dist_0 == "gumbel_r_":
        dist_kw = gumbel_r_(**dict_in)
        dist = "gumbel_r"
        samples = generate_cfd_q(
            dist_=dist, dist_kw_=dist_kw, lbound_=lbound, ubound_=ubound
        )
    elif dist_0 == "uniform_":
        dist_kw = uniform_(**dict_in)
        dist = "uniform"
        samples = generate_cfd_q(
            dist_=dist, dist_kw_=dist_kw, lbound_=lbound, ubound_=ubound
        )

    elif dist_0 == "norm_":
        dist_kw = norm_(**dict_in)
        dist = "norm"
        samples = generate_cfd_q(
            dist_=dist, dist_kw_=dist_kw, lbound_=lbound, ubound_=ubound
        )

    elif dist_0 == "lognorm_":
        dist_kw = lognorm_(**dict_in)
        dist = "lognorm"
        samples = generate_cfd_q(
            dist_=dist, dist_kw_=dist_kw, lbound_=lbound, ubound_=ubound
        )

    elif dist_0 == "lognorm_mod_":
        dist_kw = lognorm_(**dict_in)
        dist = "lognorm"
        samples = generate_cfd_q(
            dist_=dist, dist_kw_=dist_kw, lbound_=lbound, ubound_=ubound
        )
        samples = 1 - samples

    elif dist_0 == "constant_":
        # print(num_samples, lbound, ubound, np.average(lbound))
        samples = np.full((num_samples,), np.average([lbound, ubound]))

    else:
        try:
            dict_in.pop("dist")
            dict_in.pop("ubound")
            dict_in.pop("lbound")
            samples = generate_cfd_q(
                dist_=dist, dist_kw_=dict_in, lbound_=lbound, ubound_=ubound
            )
        except AttributeError:
            raise ValueError("Unknown distribution type {}.".format(dist))

    samples[samples == np.inf] = ubound
    samples[samples == -np.inf] = lbound

    if "permanent" in dict_in:
        samples += dict_in["permanent"]

    np.random.shuffle(samples)

    return samples


def dict_unflatten(dict_in: dict) -> dict:

    dict_out = dict()

    for k in list(dict_in.keys()):
        if ":" in k:
            k1, k2 = k.split(":")

            if k1 in dict_out:
                dict_out[k1][k2] = dict_in[k]
            else:
                dict_out[k1] = dict(k2=dict_in[k])

    return dict_out


def dict_flatten(dict_in: dict) -> dict:
    """Converts two levels dict to single level dict. Example input and output see _test_dict_flatten.
    :param dict_in: Any two levels (or less) dict.
    :return dict_out: Single level dict.
    """

    dict_out = dict()

    for k in list(dict_in.keys()):
        if isinstance(dict_in[k], dict):
            for kk, vv in dict_in[k].items():
                dict_out[f"{k}:{kk}"] = vv
        else:
            dict_out[k] = dict_in[k]

    return dict_out


def _test_dict_flatten():
    x = dict(A=dict(a=0, b=1), B=dict(c=2, d=3))
    y_expected = {"A:a": 0, "A:b": 1, "B:c": 2, "B:d": 3}
    y = dict_flatten(x)
    assert y == y_expected


def main(x: dict, num_samples: int) -> pd.DataFrame:
    """Generates samples based upon prescribed distribution types.

    :param x: description of distribution function.
    :param num_samples: number of samples to be produced.
    :return df_out:
    """

    dict_out = dict()

    for k, v in x.items():

        if isinstance(v, float) or isinstance(v, int) or isinstance(v, np.float):
            dict_out[k] = np.full((num_samples,), v, dtype=float)

        elif isinstance(v, str):
            dict_out[k] = np.full(
                (num_samples,), v, dtype=np.dtype("U{:d}".format(len(v)))
            )

        elif isinstance(v, np.ndarray) or isinstance(v, list):
            dict_out[k] = list(np.full((num_samples, len(v)), v, dtype=float))

        elif isinstance(v, dict):
            if "dist" in v:
                try:
                    dict_out[k] = random_variable_generator(v, num_samples)
                except KeyError:
                    raise ("Missing parameters in input variable {}.".format(k))
            elif "ramp" in v:
                s_ = StringIO(v["ramp"])
                d_ = pd.read_csv(
                    s_,
                    names=["x", "y"],
                    dtype=float,
                    skip_blank_lines=True,
                    skipinitialspace=True,
                )
                t_ = d_.iloc[:, 0]
                v_ = d_.iloc[:, 1]
                f_interp = interp1d(t_, v_, bounds_error=False, fill_value=0)
                dict_out[k] = np.full((num_samples,), f_interp)
            else:
                raise ValueError("Unknown input data type for {}.".format(k))
        else:
            raise TypeError("Unknown input data type for {}.".format(k))

    dict_out["index"] = np.arange(0, num_samples, 1)

    df_out = pd.DataFrame.from_dict(dict_out, orient="columns")

    return df_out


def _test_random_variable_generator():
    x = dict(v=np.pi)
    y = main(x, 1000)
    assert len(y["v"].values) == 1000
    assert all([v == np.pi for v in y["v"].values])

    x = dict(v="hello world.")
    y = main(x, 1000)
    assert len(y["v"].values) == 1000
    assert all([v == "hello world." for v in y["v"].values])

    x = dict(v=[0.0, 1.0, 2.0])
    y = main(x, 1000)
    assert len(y["v"].values) == 1000
    assert all([all(v == np.array([0.0, 1.0, 2.0])) for v in y["v"].values])

    x = dict(v=dict(dist="uniform_", ubound=10, lbound=-1))
    y = main(x, 1000)
    assert len(y["v"].values) == 1000
    assert np.max(y["v"].values) == 10
    assert np.min(y["v"].values) == -1
    assert abs(np.mean(y["v"].values) - (10 - 1) / 2) <= 0.00001

    x = dict(v=dict(dist="norm_", ubound=5 + 1, lbound=5 - 1, mean=5, sd=1))
    y = main(x, 1000)
    assert len(y["v"].values) == 1000
    assert np.max(y["v"].values) == 6
    assert np.min(y["v"].values) == 4
    assert abs(np.mean(y["v"].values) - 5) <= 0.00001

    x = dict(v=dict(dist="gumbel_r_", ubound=2000, lbound=50, mean=420, sd=126))
    y = main(x, 1000)
    assert len(y["v"].values) == 1000
    assert abs(np.max(y["v"].values) - 2000) <= 1
    assert abs(np.min(y["v"].values) - 50) <= 1
    assert abs(np.mean(y["v"].values) - 420) <= 1

    # todo: lognormal distribution need to be verified.
    # x = dict(
    #     v=dict(
    #         dist='lognorm_',
    #         ubound=2.5,
    #         lbound=0.00001,
    #         mean=0.00001,
    #         sd=0.25,
    #     ),
    # )
    # y = main(x, 1000)
    # assert len(y['v'].values) == 1000
    # print(np.mean(y['v'].values))
    # assert abs(np.mean(y['v'].values) - 0.5) <= 0.001


if __name__ == "__main__":
    _test_random_variable_generator()
    _test_dict_flatten()



# -*- coding: utf-8 -*-
import numpy as np
from scipy.interpolate import interp1d


def protected_steel_bs13381(
    time: np.ndarray,
    temperature_ambient: np.ndarray,
    lambda_pt_T,
    d_p: float,
    A_p: float,
    V: float,
    c_a_T,
    rho_a_T,
    time_ubound: float = 10800,
    time_step_lbound: float = 30.0,
):
    """
    This function estimates the temperature of given protected steel structural element. The calculation procedure is
    in accordance with BS EN 13381-8:2013 "Test methods for determining the contribution to the fire resistance of
    structural members - Part 8: Applied reactive protection to steel members".
    :param time:                {ndarray}   [s]         Time array incorporating temperature_ambient, forming a time-temperature curve
    :param temperature_ambient: {ndarray}   [s]         Temperature array incorporating time, forming a time-temperature curve
    :param lambda_pt_T:         {ky2T}      [W/m/K]     Thermal conductivity of the protective material (temperature dependent)
    :param d_p:                 {float}     [m]         Dry film thickness of reactive product
    :param A_p:                 {float}     [m]         Perimeter of protected
    :param V:                   {float}     [m2]        Area of steel cross section
    :param c_a_T:               {ky2T}      [J/kg/K]    Specific heat capacity of steel (temperature dependent)
    :param rho_a_T:             {ky2T}      [kg/m3]     Density of steel (temperature dependent)
    :param time_ubound:         {float}     [s]         The calculation time
    :param time_step_lbound:    {float}     [s]         The user-defined minimum time step, a smaller time step will be used with Equation E.2
    :return:
    """

    # HELPER FUNCTIONS
    # [BS EN 13381-8:2013, ANNEX E, Equation E.1]
    def _theta_at(lambda_pt, d_p, A_p, V, c_a, rho_a, theta_t, theta_at, dt):
        return (
            (lambda_pt / d_p)
            * (A_p / V)
            * (1 / c_a / rho_a)
            * (theta_t - theta_at)
            * dt
        )

    # [BS EN 13381-8:2013, ANNEX E, Equation E.2]
    def _dt(c_a, rho_a, lambda_pt, d_p, A_p, V):
        return 0.8 * (c_a * rho_a / lambda_pt * d_p) * (V / A_p)

    # [BS EN 13381-8:2013, ANNEX E, Equation E.3]
    def _lambda_pt(d_p, V, A_p, c_a, rho_a, theta_t, theta_at, dt, d_theta_at):
        return (
            d_p
            * (V / A_p)
            * c_a
            * rho_a
            * (1 / ((theta_t - theta_at) * dt))
            * d_theta_at
        )

    # [BS EN 13381-8:2013, ANNEX E, Equation E.4]
    def _theta_pt(theta_t, theta_t_, theta_at, theta_at_):
        return ((theta_t_ + theta_t) / 2.0 + (theta_at_ + theta_at) / 2.0) / 2.0

    # [BS EN 13381-8:2013, ANNEX E, Equation E.5]
    def _d_theta_at(c_a, rho_a, lambda_ave, d_p, A_p, V, theta_t, theta_at, dt):
        return (
            (1 / (c_a + rho_a))
            * (lambda_ave / d_p)
            * (A_p / V)
            * (theta_t - theta_at)
            * dt
        )

    # [BS EN 13381-8:2013, ANNEX E, Equation E.6]
    def _lambda_char(lambda_ave, K, sigma):
        return lambda_ave + K * sigma

    # Note: only equation E.1 and E.2 are currently being used as it is assumed the conductivity of protection layer
    #       is known.

    _temperature_gas = interp1d(time, temperature_ambient)

    # Instantiate output containers
    time_ = [0]
    time_rate = [0]
    temperature_steel = [temperature_ambient[0]]
    temperature_rate_steel = [0]
    conductivity_protection = [np.nan]

    # Start iterative calculation
    t = 0
    i = 0
    while t < time_ubound:
        i += 1

        # Determine conditions
        T_g = _temperature_gas(t)

        # Determine thermal properties
        c_a = c_a_T(T_g)
        rho_a = rho_a_T(T_g)
        lambda_pt = lambda_pt_T(T_g)
        conductivity_protection[i] = lambda_pt

        # Determine time step
        dt = _dt(c_a, rho_a, lambda_pt, d_p, A_p, V)
        dt = min([dt, time_step_lbound])
        time_rate.append(dt)

        # Determine current time
        time_.append(time_[i - 1] + dt)
        t = time_[i]

        # Determine temperature change rate of the steel
        lambda_ave, theta_t, theta_at = lambda_pt, T_g, temperature_steel[i - 1]
        temperature_rate_steel.append(
            _d_theta_at(c_a, rho_a, lambda_ave, d_p, A_p, V, theta_t, theta_at, dt)
        )
        temperature_steel.append(temperature_steel[i - 1] + temperature_rate_steel[i])

    # Convert lists to ndarray
    time_, time_rate = np.asarray(time_), np.asarray(time_rate)
    temperature_steel, temperature_rate_steel = (
        np.asarray(temperature_steel),
        np.asarray(temperature_rate_steel),
    )

    return time_, temperature_steel, time_rate, temperature_rate_steel



import numpy as np


def fire(time, temperature_initial):

    # INPUTS CHECK
    time = np.array(time, dtype=float)
    time[time < 0] = np.nan

    # SI UNITS -> EQUATION UNITS
    temperature_initial -= 273.15  # [K] -> [C]
    time /= 60.0  # [s] - [min]

    # CALCULATE TEMPERATURE BASED ON GIVEN TIME
    temperature = 345.0 * np.log10(time * 8.0 + 1.0) + temperature_initial
    temperature[temperature == np.nan] = temperature_initial

    # EQUATION UNITS -> SI UNITS
    time *= 60.0  # [min] -> [s]
    temperature += 273.15  # [C] -> [K]

    return temperature


if __name__ == "__main__":
    fire_time = np.arange(0, 2 * 60 * 60 + 1, 1)
    fire_temperature = fire(fire_time, 273.15 + 20)

    print(
        fire_temperature[fire_time == 5 * 60],
        fire_temperature[fire_time == 10 * 60],
        fire_temperature[fire_time == 15 * 60],
        fire_temperature[fire_time == 20 * 60],
        fire_temperature[fire_time == 25 * 60],
    )



# -*- coding: utf-8 -*-

from sfeprapy.mcs0.mcs0_calc import _test_teq_phi as test_teq_phi
from sfeprapy.mcs0.mcs0_calc import _test_standard_case as test_mcs0_test_standard_case

test_teq_phi()
test_mcs0_test_standard_case()



#!/usr/bin/env python

import os
from codecs import open  # To use a consistent encoding

import setuptools

import sfeprapy

with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "README.md")) as f:
    long_description = f.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="sfeprapy",
    version=sfeprapy.__version__,
    description="Structural Fire Engineering - Probabilistic Reliability Assessment (Equivalent Time Exposure)",
    author="Ian Fu",
    author_email="fuyans@gmail.com",
    url="https://github.com/fsepy/sfeprapy",
    download_url="https://github.com/fsepy/sfeprapy/archive/master.zip",
    keywords=["fire safety", "structural fire engineering"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=[
        "sfeprapy",
        "sfeprapy.dat",
        "sfeprapy.mcs0",
        "sfeprapy.mcs1",
        "sfeprapy.func",
        "sfeprapy.dist_fit",
    ],
    install_requires=requirements,
    include_package_data=True,
    entry_points={"console_scripts": ["sfeprapy=sfeprapy.cli:main"]},
)
